# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated (+).

"""Total variation filter tool."""

from ..gtk.customwidgets import VBox, RadioButtons, HScaleSpinButton
from ..toolmanager import BaseToolWindow
from skimage.restoration import denoise_tv_chambolle, denoise_tv_bregman

class TotalVariationFilterTool(BaseToolWindow):
  """Total variation filter tool class."""

  _action_ = "Applying total variation filter..."

  _help_ = """Total variation denoising.
Given a noisy image f, find an image u with less total variation than f under the constraint that u remains similar to f. This can be expressed as the Rudin–Osher–Fatemi (ROF) minimization problem:

    minmize \u03a3r |\u2207u(r)|+(\u03bb/2)[f(r)-u(r)]²

where the weight 1/\u03bb controls denoising (the larger the weight, the stronger the denoising at the expense of image fidelity).
The minimization can either be performed with the Chambolle or Split Bregman algorithms.
Total variation denoising tends to produce cartoon-like (piecewise-constant) images."""

  _onthefly_ = False # This transformation can not be applied on the fly.

  def open(self, image):
    """Open tool window for image 'image'."""
    if not super().open(image, "Total variation filter"): return False
    wbox = VBox()
    self.window.add(wbox)
    self.widgets.weightscale = HScaleSpinButton(.1, 0., 1., .001, digits = 3, length = 480)
    wbox.pack(self.widgets.weightscale.layout2("Weight:"))
    self.widgets.algobuttons = RadioButtons(("Chambolle", "Chambolle"), ("Bregman", "Split Bregman"))
    wbox.pack(self.widgets.algobuttons.hbox(prepend = "Algorithm:"))
    wbox.pack(self.tool_control_buttons())
    self.start(identity = False)
    return True

  def get_params(self):
    """Return tool parameters."""
    return self.widgets.weightscale.get_value(), self.widgets.algobuttons.get_selected()

  def set_params(self, params):
    """Set tool parameters 'params'."""
    weight, algorithm = params
    self.widgets.weightscale.set_value(weight)
    self.widgets.algobuttons.set_selected(algorithm)

  def run(self, params):
    """Run tool for parameters 'params'."""
    weight, algorithm = params
    if weight <= 0.: return params, False
    if algorithm == "Chambolle":
      self.image.rgb = denoise_tv_chambolle(self.reference.rgb, channel_axis = 0, weight = weight)
    else:
      self.image.rgb = denoise_tv_bregman(self.reference.rgb, channel_axis = 0, weight = 1./(2.*weight))
    return params, True

  def operation(self, params):
    """Return tool operation string for parameters 'params'."""
    weight, algorithm = params
    return f"TotalVariationFilter(weight = {weight:.3f}, algorithm = {algorithm})"



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updatedi (+).

"""Blend tool."""

from ..gtk.customwidgets import Align, Label, HBox, VBox, Grid, CheckButton, HScale
from ..misc.imagechooser import ImageChooser
from ..toolmanager import BaseToolWindow
import numpy as np

class BlendTool(BaseToolWindow):
  """Blend tool window class."""

  _action_ = "Blending images..."

  _help_ = """Blend the current image IMG with an other image BLEND of your choice:

    OUT = f*BLEND+(1-f)*IMG

The mixing factor f can be tuned independently for each red/blue/green channel (untick the "bind RGB channels" checkbox).
If a checkbox "zero is transparent" is ticked, the zero pixels in that channel of BLEND are treated as transparent (not blended with the current image)."""

  def open(self, image):
    """Open tool window for image 'image'."""
    if not super().open(image, "Blend images"): return False
    wbox = VBox()
    self.window.add(wbox)
    wbox.pack("Choose image to blend with:")
    self.widgets.chooser = ImageChooser(self.app, self.window, wbox, callback = lambda row, image: self.update("image"))
    self.message = Label(" ")
    wbox.pack(self.message)
    hbox = HBox()
    wbox.pack(hbox)
    hbox.pack("Mixing factors:", expand = True, fill = True)
    self.widgets.bindbutton = CheckButton(label = "Bind RGB channels")
    self.widgets.bindbutton.set_active(True)
    self.widgets.bindbutton.connect("toggled", lambda button: self.update(0))
    hbox.pack(self.widgets.bindbutton)
    grid = Grid()
    wbox.pack(grid)
    self.widgets.mixingscales = []
    self.widgets.zerobuttons = []
    for channel, label in ((0, "Red:"), (1, "Green:"), (2, "Blue:")):
      mixingscale = HScale(0., -1., 2., .01, digits = 2, marks = [-1., 0., 1., 2.], length = 320)
      mixingscale.channel = channel
      mixingscale.connect("value-changed", lambda scale: self.update(scale.channel))
      self.widgets.mixingscales.append(mixingscale)
      zerobutton = CheckButton(label = "Zero is transparent")
      zerobutton.channel = channel
      zerobutton.connect("toggled", lambda button: self.update(button.channel))
      self.widgets.zerobuttons.append(zerobutton)
      grid.attach(Label(label, halign = Align.END), 0, channel)
      grid.attach(mixingscale, 1, channel)
      grid.attach(zerobutton, 2, channel)
    wbox.pack(self.tool_control_buttons())
    self.start(identity = True)
    return True

  def get_params(self):
    """Return tool parameters."""
    row = self.widgets.chooser.get_selected_row()
    mixings = tuple(self.widgets.mixingscales[channel].get_value() for channel in range(3))
    zeros = tuple(self.widgets.zerobuttons[channel].get_active() for channel in range(3))
    return row, mixings, zeros

  def set_params(self, params):
    """Set tool parameters 'params'."""
    row, mixings, zeros = params
    self.widgets.chooser.set_selected_row(row)
    for channel in range(3):
      self.widgets.mixingscales[channel].set_value_block(mixings[channel])
      self.widgets.zerobuttons[channel].set_active_block(zeros[channel])
    if mixings[1] != mixings[0] or mixings[2] != mixings[0]: self.widgets.bindbutton.set_active_block(False)
    if zeros[1] != zeros[0] or zeros[2] != zeros[0]: self.widgets.bindbutton.set_active_block(False)
    self.update("all")

  def run(self, params):
    """Run tool for parameters 'params'."""
    row, mixings, zeros = params
    if row < 0: return params, False
    selection = self.widgets.chooser.get_image(row)
    if selection.size() != self.reference.size():
      self.set_message("<span foreground='red'>Can not blend images with different sizes.</span>")
      return params, False
    self.set_message()
    for channel in range(3):
      mixing = mixings[channel]
      zero = zeros[channel]
      blended = mixing*selection.rgb[channel]+(1.-mixing)*self.reference.rgb[channel]
      if zero:
        self.image.rgb[channel] = np.where(selection.rgb[channel] > 0., blended, self.reference.rgb[channel])
      else:
        self.image.rgb[channel] = blended
    return params, True

  def operation(self, params):
    """Return tool operation string for parameters 'params'."""
    row, mixings, zeros = params
    if row < 0: return None
    operation = f"Blend({self.widgets.chooser.get_image_tag(row)}"
    for channel in range(3):
      key = ["R", "G", "B"][channel]
      decoration = "'" if zeros[channel] else ""
      operation += f", {key}{decoration} = {mixings[channel]:.2f}"
    operation += ")"
    return operation

  # Update widgets.

  def update(self, changed):
    """Update widgets on change of 'changed'."""
    if changed in [0, 1, 2]:
      if self.widgets.bindbutton.get_active():
        mixing = self.widgets.mixingscales[changed].get_value()
        transparent = self.widgets.zerobuttons[changed].get_active()
        for channel in range(3):
          self.widgets.mixingscales[channel].set_value_block(mixing)
          self.widgets.zerobuttons[channel].set_active_block(transparent)
    self.reset_polling(self.get_params()) # Expedite main window update.

  def set_message(self, message = " "):
    """Set message 'message'."""
    self.message.set_markup(message)



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated.

"""Contrast Limited Adaptive Histogram Equalization (CLAHE) tool."""

from ..gtk.customwidgets import HBox, VBox, CheckButton, RadioButtons, HScaleSpinButton
from ..toolmanager import BaseToolWindow
from ...imageprocessing import imageprocessing
from skimage.exposure import equalize_adapthist

class CLAHETool(BaseToolWindow):
  """Contrast Limited Adaptive Histogram Equalization (CLAHE) tool class."""

  _action_ = "Running Contrast Limited Adaptive Histogram Equalization (CLAHE)..."

  _onthefly_ = False # This transformation can not be applied on the fly.

  def open(self, image):
    """Open tool window for image 'image'."""
    if not super().open(image, "Contrast Limited Adaptive Histogram Equalization (CLAHE)"): return False
    wbox = VBox()
    self.window.add(wbox)
    self.widgets.highlightsbutton = CheckButton(label = "Protect highlights")
    self.widgets.highlightsbutton.set_sensitive(False)
    self.widgets.channelbuttons = RadioButtons(("V", "HSV value"), ("L", "Luma"))
    self.widgets.channelbuttons.buttons["L"].connect("toggled", \
      lambda button: self.widgets.highlightsbutton.set_sensitive(self.widgets.channelbuttons.get_selected() == "L"))
    wbox.pack(self.widgets.channelbuttons.hbox(prepend = "Channel:", append = self.widgets.highlightsbutton))
    self.widgets.sizebutton = HScaleSpinButton(12.5, 1., 99., .01, digits = 2, length = 480)
    wbox.pack(self.widgets.sizebutton.layout2("Kernel size (% image width and height):"))
    self.widgets.clipscale = HScaleSpinButton(.01, 0., .1, .0001, digits = 4, length = 480)
    wbox.pack(self.widgets.clipscale.layout2("Clip limit:"))
    wbox.pack(self.tool_control_buttons())
    self.start(identity = False)
    return True

  def get_params(self):
    """Return tool parameters."""
    return self.widgets.channelbuttons.get_selected(), self.widgets.sizebutton.get_value(), self.widgets.clipscale.get_value(), \
           self.widgets.highlightsbutton.get_active(), imageprocessing.get_rgb_luma()

  def set_params(self, params):
    """Set tool parameters 'params'."""
    channel, size, clip, highlights, rgbluma = params
    self.widgets.channelbuttons.set_selected(channel)
    self.widgets.sizebutton.set_value(size)
    self.widgets.clipscale.set_value(clip)
    self.widgets.highlightsbutton.set_active(highlights)

  def run(self, params):
    """Run tool for parameters 'params'."""
    channel, size, clip, highlights, rgbluma = params
    if size <= 0. or clip <= 0.: return params, False
    width, height = self.reference.size()
    kwidth = max(int(round(size*width/100.)), 3)
    kheight = max(int(round(size*height/100.)), 3)
    nbins = min(2**self.app.get_color_depth(), 1024)
    print(f"Using {nbins} bins...")
    reference = self.reference.clone()
    reference.clip() # Clip before CLAHE.
    if channel == "V":
      self.image.set_image(equalize_adapthist(reference.rgbf_view(), kernel_size = (kheight, kwidth), clip_limit = clip, nbins = nbins), channels = -1)
    else:
      ref = reference.luma()
      img = equalize_adapthist(ref, kernel_size = (kheight, kwidth), clip_limit = clip, nbins = nbins)
      self.image.copy_image_from(reference)
      self.image.scale_pixels(ref, img)
      if highlights: self.image.protect_highlights()
    return params, True

  def operation(self, params):
    """Return tool operation string for parameters 'params'."""
    channel, size, clip, highlights, rgbluma = params
    optstring = ""
    if channel == "L":
      channel = f"L({rgbluma[0]:.2f}, {rgbluma[1]:.2f}, {rgbluma[2]:.2f})"
      if highlights: optstring = ", protect highlights"
    return f"CLAHE(channel = {channel}, size = {size:.2f}%, clip = {clip:.4f}{optstring})"



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated (+).

"""Color saturation tool."""

from ..gtk.customwidgets import Align, Label, HBox, VBox, Grid, CheckButton, RadioButtons, HScaleSpinButton
from ..base import FigureCanvas
from ..toolmanager import BaseToolWindow
import numpy as np
import matplotlib.colors as colors
import matplotlib.ticker as ticker
from matplotlib.figure import Figure
from scipy.interpolate import interp1d, splrep, splev

class ColorSaturationTool(BaseToolWindow):
  """Color saturation tool class."""

  _action_ = "Tuning color saturation..."

  _help_ = """Tune saturation in the HSV color space.
The different hues (red, yellow, green, cyan, blue, magenta) can be (de)saturated jointly (if the "bind hues" checkbox is ticked) or independently.
The saturation of each hue is updated as

  sat \u2192 sat+\u0394  if model is "\u0394Sat"

or as

  sat \u2192 mtf(sat, (1-\u0394)/2)  if model is "MidSatStretch",

where \u0394\u2208[-1, 1] is the value of the corresponding spin button and mtf(x, m) = (m-1)x/((2m-1)x-m) is the midtone stretch function.
\u0394 is interpolated between different hues using nearest neighbor, linear or cubic interpolation. The interpolated \u0394 is plotted as a dashed line on the HSV wheel."""

  def open(self, image):
    """Open tool window for image 'image'."""
    if not super().open(image, "Color saturation"): return False
    self.reference.hsv = self.reference.rgb_to_hsv()
    wbox = HBox(spacing = 16)
    self.window.add(wbox)
    self.widgets.fig = Figure(figsize = (6., 6.))
    canvas = FigureCanvas(self.widgets.fig)
    canvas.set_size_request(300, 300)
    wbox.pack(canvas)
    vbox = VBox()
    wbox.pack(vbox, expand = True, fill = True)
    grid = Grid(row_spacing = 2)
    vbox.pack(grid)
    self.widgets.modelbuttons = RadioButtons(("DeltaSat", "\u0394Sat"), ("MidSatStretch", "MidSat stretch"))
    self.widgets.modelbuttons.connect("toggled", lambda button: self.update(-2))
    grid.attach(Label("Model:", halign = Align.END), 0, 0)
    grid.attach(self.widgets.modelbuttons.hbox(), 1, 0)
    self.widgets.bindbutton = CheckButton(label = "Bind hues", halign = Align.END)
    self.widgets.bindbutton.set_active(True)
    self.widgets.bindbutton.connect("toggled", lambda button: self.update(0))
    grid.attach(self.widgets.bindbutton, 2, 0)
    self.widgets.satscales = []
    for hid, label in ((0, "Red:"), (1, "Yellow:"), (2, "Green:"), (3, "Cyan:"), (4, "Blue:"), (5, "Magenta:")):
      satscale = HScaleSpinButton(0., -1., 1., .001, digits = 3, length = 320)
      satscale.hid = hid
      satscale.connect("value-changed", lambda scale: self.update(scale.hid))
      self.widgets.satscales.append(satscale)
      grid.attach(Label(label, halign = Align.END), 0, hid+1)
      grid.attach(satscale.layout1(), 1, hid+1, width = 2)
    self.widgets.interbuttons = RadioButtons(("nearest", "Nearest"), ("linear", "Linear"), ("cubic", "Cubic"))
    self.widgets.interbuttons.connect("toggled", lambda button: self.update(-1))
    grid.attach(Label("Interpolation:", halign = Align.END), 0, 7)
    grid.attach(self.widgets.interbuttons.hbox(), 1, 7, width = 2)
    vbox.pack(self.tool_control_buttons())
    self.plot_hsv_wheel()
    self.outofrange = self.reference.is_out_of_range() # Is the reference image out-of-range ?
    if self.outofrange: print("Reference image is out-of-range...")
    self.start(identity = not self.outofrange) # If so, the color saturation tool will clip the image whatever the parameters
    return True

  def get_params(self):
    """Return tool parameters."""
    model = self.widgets.modelbuttons.get_selected()
    psat = tuple(self.widgets.satscales[hid].get_value() for hid in range(6)) # As a tuple because parameters need to be copied.
    interpolation = self.widgets.interbuttons.get_selected()
    return model, psat, interpolation

  def set_params(self, params):
    """Set tool parameters 'params'."""
    model, psat, interpolation = params
    psat = np.array(psat)
    self.widgets.modelbuttons.set_selected_block(model)
    for hid in range(6): self.widgets.satscales[hid].set_value_block(psat[hid])
    if np.any(psat != psat[0]): self.widgets.bindbutton.set_active_block(False)
    self.widgets.interbuttons.set_selected_block(interpolation)
    self.update(-1)

  def run(self, params):
    """Run tool for parameters 'params'."""
    model, psat, interpolation = params
    psat = np.array(psat)
    if not self.outofrange and np.all(psat == 0.): return params, False
    hsv = self.reference.hsv.copy()
    hue = hsv[:, :, 0]
    sat = hsv[:, :, 1]
    delta = self.interpolate(hue, psat, interpolation)
    if model == "DeltaSat":
      sat += delta
    else:
      midsat = np.clip(.5*(1.-delta), .005, .995)
      sat = (midsat-1.)*sat/((2.*midsat-1.)*sat-midsat)
    hsv[:, :, 1] = np.clip(sat, 0., 1.)
    self.image.set_hsv_image(hsv)
    return params, True

  def operation(self, params):
    """Return tool operation string for parameters 'params'."""
    model, psat, interpolation = params
    tags = ["R", "Y", "G", "C", "B", "M"]
    operation = f"ColorSaturation(model = {model}, "
    for hid in range(6):
      operation += f"{tags[hid]} = {psat[hid]:.3f}, "
    operation += f" interpolation = {interpolation})"
    return operation

  # Interpolate saturation parameters.

  def interpolate(self, hue, psat, interpolation):
    """Interpolate the saturation parameter psat[RYGCBM] for arbitrary hues."""
    if np.all(psat == psat[0]):
      return np.full_like(hue, psat[0]) # Short-cut if the saturation parameter is the same for RYGCBM.
    hsat = np.linspace(0., 1., 7)
    psat = np.append(psat, psat[0])
    if interpolation == "nearest":
      fsat = interp1d(hsat, psat, kind = "nearest")
    else:
      k = 3 if interpolation == "cubic" else 1
      tck = splrep(hsat, psat, k = k, per = True) # Enforce periodic boundary conditions.
      def fsat(x): return np.clip(splev(x, tck), -1., 1.)
    return fsat(hue)

  # Plot HSV wheel.

  def plot_hsv_wheel(self):
    """Plot HSV wheel."""
    ax = self.widgets.fig.add_axes([.1, .1, .8, .8], projection = "polar")
    self.widgets.fig.satax = ax
    ax.patch.set_alpha(0.)
    ax2 = ax.figure.add_axes(ax.get_position(), projection = "polar", zorder = -3)
    ax2.axis("off")
    ax2.set_ylim(0., 1.)
    rho = np.linspace(1., 1.2, 32)
    phi = np.linspace(0., 2.*np.pi, 256)
    RHO, PHI = np.meshgrid(rho, phi)
    h = np.ravel(PHI/(2.*np.pi))
    s = np.ones_like(h)
    v = s
    hsv = np.column_stack((h, s, v))
    rgb = colors.hsv_to_rgb(hsv)
    ax2.scatter(PHI, RHO, c = rgb, clip_on = False)
    hue = np.linspace(0., 5.*np.pi/3., 6)
    ax.satpoints, = ax.plot(hue, np.zeros_like(hue), "ko", ms = 8)
    ax.set_xticks(hue, labels = ["R", "Y", "G", "C", "B", "M"])
    ax.set_ylim([-1., 1.])
    ax.yaxis.set_major_locator(ticker.LinearLocator(5))
    ax.yaxis.set_major_formatter(ticker.FormatStrFormatter("%.2f"))
    hue = np.linspace(0., 2.*np.pi, 128)
    ax.satcurve, = ax.plot(hue, np.zeros_like(hue), "k--")

  # Update widgets.

  def update(self, changed):
    """Update widgets on change of 'changed'."""
    if changed >= 0:
      if self.widgets.bindbutton.get_active():
        psat = self.widgets.satscales[changed].get_value()
        for hid in range(6):
          self.widgets.satscales[hid].set_value_block(psat)
    if changed >= -1: self.update_hsv_wheel()
    self.reset_polling(self.get_params()) # Expedite main window update.

  def update_hsv_wheel(self):
    """Update HSV wheel."""
    model, psat, interpolation = self.get_params()
    psat = np.array(psat)
    ax = self.widgets.fig.satax
    ax.satpoints.set_ydata(psat)
    ax.satcurve.set_ydata(self.interpolate(ax.satcurve.get_xdata()/(2.*np.pi), psat, interpolation))
    pmin = psat.min()
    pmax = psat.max()
    if np.all(psat == psat[0]) or pmax-pmin > .25:
      ymin = -1.
      ymax =  1.
    else:
      ymin = max(pmin-.125, -1.)
      ymax = min(pmax+.125,  1.)
      if ymax-ymin < .2:
        if ymax ==  1.:
          ymin =  .8
        elif ymin == -1.:
          ymax = -.8
    ax.set_ylim(ymin, ymax)
    self.widgets.fig.canvas.draw_idle()



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated.

"""Image chooser widget."""

import os
import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GObject
from ..gtk.customwidgets import HBox, VBox, HButtonBox, ScrolledBox, Button
from ..gtk.filechoosers import ImageFileChooserDialog
from ..base import ErrorDialog

class ImageChooser():
  """Image chooser widget class."""

  def __init__(self, app, window, vbox, tabkey = "Selection", callback = None, last = False):
    """Add an image chooser treeview in VBox 'vbox' of window 'window' of app 'app'. Display the selected image in tab 'tabkey' of the main
       window (if not None). 'callback(row, image)' is an optional method called upon selection change, with 'image' the selected
       image on row 'row' of the treeview. If 'last' is False, the last image/operation (the current image) is not included in the treeview."""
    self.app = app
    self.window = window
    self.tabkey = tabkey
    self.newtab = tabkey not in self.app.mainwindow.get_keys() # Is this a new tab ?
    self.callback = callback
    self.nfiles = 0
    self.nimages = 0
    self.imagestore = Gtk.ListStore(int, str, GObject.TYPE_PYOBJECT)
    for operation, image, frame in self.app.operations[:None if last else -1]:
      self.nimages += 1
      self.imagestore.append([self.nimages, operation, image])
    scrolled = ScrolledBox(480, 200)
    vbox.pack(scrolled, expand = True, fill = True)
    self.treeview = Gtk.TreeView(model = self.imagestore, search_column = -1)
    scrolled.add(self.treeview)
    renderer = Gtk.CellRendererText()
    renderer.set_property("xalign", 1.)
    column = Gtk.TreeViewColumn("#", renderer, text = 0)
    column.set_alignment(.5)
    column.set_expand(False)
    self.treeview.append_column(column)
    renderer = Gtk.CellRendererText()
    renderer.set_property("xalign", 0.)
    column = Gtk.TreeViewColumn("Operation", renderer, text = 1)
    column.set_alignment(0.)
    column.set_expand(True)
    self.treeview.append_column(column)
    self.selection = self.treeview.get_selection()
    self.selection.set_mode(Gtk.SelectionMode.SINGLE)
    self.selection.connect("changed", lambda selection: self.update())
    self.buttonbox = HButtonBox()
    vbox.pack(self.buttonbox)
    self.filebutton = Button(label = "Add file")
    self.filebutton.connect("clicked", self.load_file)
    self.buttonbox.pack(self.filebutton)

  def load_file(self, *args, **kwargs):
    """Open file dialog and load an extra image file."""
    filename = ImageFileChooserDialog(self.window, Gtk.FileChooserAction.OPEN, preview = True)
    if filename is None: return
    try:
      image = self.app.ImageClass()
      image.load(filename)
    except Exception as err:
      ErrorDialog(self.window, str(err))
      return
    self.nfiles += 1
    self.nimages += 1
    basename = os.path.basename(filename)
    image.meta["imchooser.row"] = self.nimages
    image.meta["imchooser.file"] = basename
    image.meta["imchooser.tag"] = f"file = '{basename}'"
    image.meta["description"] = basename
    self.imagestore.append([self.nimages, f"Load('{basename}')", image])

  def get_image(self, row):
    """Get image on row 'row'."""
    return self.imagestore[row][2] if row >= 0 and row < self.nimages else None

  def get_image_tag(self, row):
    """Get tag of image on row 'row'."""
    return self.imagestore[row][2].meta.get("imchooser.tag", f"image = #{row+1}") if row >= 0 and row < self.nimages else ""

  def get_selected_row(self):
    """Return selected row."""
    model, list_iter = self.selection.get_selected()
    return model[list_iter][0]-1 if list_iter is not None else -1

  def set_selected_row(self, row):
    """Set selected row 'row'."""
    if row < 0: row = self.nimages+row
    if row >= 0 and row < self.nimages: self.treeview.set_cursor(row)

  def get_selected_image(self):
    """Return selected image."""
    model, list_iter = self.selection.get_selected()
    return model[list_iter][2] if list_iter is not None else None

  def get_selected_row_and_image(self):
    """Return selected row and image."""
    model, list_iter = self.selection.get_selected()
    return (model[list_iter][0]-1, model[list_iter][2]) if list_iter is not None else (None, None)

  def get_images_list(self):
    """Return the list of images in the treeview."""
    return [image for n, operation, image in self.imagestore]

  def update(self):
    """Update main window selection tab."""
    row, image = self.get_selected_row_and_image()
    if self.tabkey is not None:
      if image is None:
        self.app.mainwindow.delete_image(self.tabkey, force = self.newtab, failsafe = True)
      else:
        self.app.mainwindow.update_image(self.tabkey, image, create = self.newtab)
    if self.callback is not None: self.callback(row, image)

  def lock(self):
    """Lock treeview."""
    self.filebutton.set_sensitive(False)

  def unlock(self):
    """Unlock treeview."""
    self.filebutton.set_sensitive(True)



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated (+).

"""Log window."""

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
from .gtk.customwidgets import VBox, ScrolledBox, HButtonBox, Button, TextView
from .base import BaseWindow, Container

class LogWindow(BaseWindow):
  """Log window class."""

  def open(self):
    """Open log window."""
    if self.opened: return
    if self.app.get_basename() is None: return
    self.opened = True
    self.window = Gtk.Window(title = f"Logs for {self.app.get_basename()}", border_width = 16)
    self.window.connect("delete-event", self.close)
    self.window.set_size_request(480, 360)
    self.widgets = Container()
    wbox = VBox()
    self.window.add(wbox)
    self.widgets.scrolled = ScrolledBox(-1, -1)
    wbox.pack(self.widgets.scrolled, expand = True, fill = True)
    self.widgets.textview = TextView()
    self.widgets.scrolled.add(self.widgets.textview)
    hbox = HButtonBox()
    wbox.pack(hbox)
    self.widgets.copybutton = Button(label = "Copy")
    self.widgets.copybutton.connect("clicked", self.widgets.textview.copy_to_clipboard)
    hbox.pack(self.widgets.copybutton)
    self.widgets.closebutton = Button(label = "Close")
    self.widgets.closebutton.connect("clicked", self.close)
    hbox.pack(self.widgets.closebutton)
    self.update()
    self.window.show_all()

  def close(self, *args, **kwargs):
    """Close log window."""
    if not self.opened: return
    self.window.destroy()
    self.opened = False
    del self.widgets

  def update(self):
    """Update log window."""
    if not self.opened: return
    self.widgets.textview.set_text(self.app.logs())



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated.

"""Image light curve window."""

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
from .gtk.customwidgets import HBox, VBox, Button, RadioButtons
from .base import BaseWindow, FigureCanvas, BaseToolbar, Container
from .misc.imagechooser import ImageChooser
from matplotlib.figure import Figure
import numpy as np

class LightCurveWindow(BaseWindow):
  """Image light curve window class."""

  key = None # Registered key of the image tab.

  def open(self, image, key, tab):
    """Open light curve window for image 'image' of tab 'tab' with key 'key'."""
    if self.opened: self.close()
    self.key = key
    self.opened = True
    self.image = image
    self.window = Gtk.Window(title = f"Light curve of tab '{tab}'", transient_for = self.app.mainwindow.window, destroy_with_parent = True, border_width = 16)
    self.window.connect("delete-event", self.close)
    self.widgets = Container()
    wbox = VBox()
    self.window.add(wbox)
    fbox = VBox(spacing = 0)
    wbox.pack(fbox, expand = True, fill = True)
    self.widgets.fig = Figure(figsize = (8., 8.), layout = "constrained")
    canvas = FigureCanvas(self.widgets.fig)
    canvas.set_size_request(480, 480)
    fbox.pack(canvas, expand = True, fill = True)
    toolbar = BaseToolbar(canvas, self.widgets.fig)
    fbox.pack(toolbar)
    self.widgets.fig.ax = self.widgets.fig.add_subplot(111)
    self.widgets.channelbuttons = RadioButtons(("V", "HSV value"), ("L", "Luma"), ("Y", "Luminance Y"), ("L*", "Lightness L*"))
    self.widgets.channelbuttons.set_selected("L*")
    self.widgets.channelbuttons.connect("toggled", self.update)
    wbox.pack(self.widgets.channelbuttons.hbox(prepend = "Channel:"))
    wbox.pack("Reference:")
    self.widgets.chooser = ImageChooser(self.app, self.window, wbox, tabkey = None, callback = self.update, last = True)
    self.widgets.closebutton = Button(label = "Close")
    self.widgets.closebutton.connect("clicked", self.close)
    self.widgets.chooser.buttonbox.pack(self.widgets.closebutton)
    self.widgets.chooser.set_selected_row(0)
    self.window.show_all()

  def update(self, *args, **kwargs):
    reference = self.widgets.chooser.get_selected_image()
    channel = self.widgets.channelbuttons.get_selected()
    if channel == "V":
      ref = reference.value()
      img = self.image.value()
      label = "HSV value"
    elif channel == "L":
      ref = reference.luma()
      img = self.image.luma()
      label = "luma"
    elif channel == "Y":
      ref = reference.srgb_luminance()
      img = self.image.srgb_luminance()
      label = "luminance"
    else:
      ref = reference.srgb_lightness()/100.
      img = self.image.srgb_lightness()/100.
      label = "lightness"
    ax = self.widgets.fig.ax
    ax.clear()
    ax.set_xlim(0., 1.)
    ax.set_xlabel("Reference "+label)
    ax.set_ylim(0., 1.)
    ax.set_ylabel("Image "+label)
    ax.plot([0., 1.], [0., 1.], color = "gray", linestyle = ":", linewidth = 1., zorder = -3)
    if ref.shape != img.shape:
      ax.plot([1., 0.], [0., 1.], color = "gray", linestyle = ":", linewidth = 1., zorder = -3)
      ax.text(.5, .5, "Image sizes do not match", color = "red", fontsize = 16, ha = "center", va = "center")
    else:
      maxsize = np.max(img.shape)
      n = int(np.ceil(maxsize/1024))
      ax.plot(np.ravel(ref[::n, ::n]), np.ravel(img[::n, ::n]), ".")
      ax.grid()
    self.widgets.fig.canvas.draw_idle()

  def close(self, *args, **kwargs):
    """Close light curve window.
       If the kwarg 'key' is not None, close only if self.key = key."""
    if not self.opened: return
    key = kwargs.get("key", None)
    if key is not None and self.key != key: return
    self.window.destroy()
    self.opened = False
    self.key = None
    del self.widgets



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17

"""Histogram stretch functions."""

import numpy as np

def midtone_stretch_function(x_, params):
  """Return the midtone stretch function f(x_) for parameters 'params' = (shadow, midtone, highlight, low, high)."""
  shadow, midtone, highlight, low, high = params
  midtone = (midtone-shadow)/(highlight-shadow)
  x = np.clip(x_, shadow, highlight)
  x = np.interp(x, (shadow, highlight), (0., 1.))
  y = (midtone-1.)*x/((2.*midtone-1.)*x-midtone)
  return np.interp(y, (low, high), (0., 1.))

def blackpoint_stretch_function(x_, params):
  """Return the linear black point stretch function f(x_) for parameters 'params' = (shadow, )."""
  shadow, = params
  x = np.clip(x_, shadow, 1.)
  return np.interp(x, (shadow, 1.), (0., 1.))

def arcsinh_stretch_function(x_, params):
  """Return the arcsinh stretch function f(x_) for parameters 'params' = (shadow, stretch)."""
  shadow, stretch = params
  x = np.clip(x_, shadow, 1.)
  x = np.interp(x, (shadow, 1.), (0., 1.))
  if abs(stretch) < 1.e-6: # Identity.
    return x
  else:
    return np.arcsinh(stretch*x)/np.arcsinh(stretch)

def ghyperbolic_stretch_function(x_, params):
  """Return the generalized hyperbolic stretch function f(x_) for parameters 'params' = (log(D+1), B, SYP, SPP, HPP, inverse).
     See: https://ghsastro.co.uk/.
     Code adapted from https://github.com/mikec1485/GHS/blob/main/src/scripts/GeneralisedHyperbolicStretch/lib/GHSStretch.js."""
  logD1, B, SYP, SPP, HPP, inverse = params
  D = np.exp(logD1)-1.
  x = np.clip(x_, 0., 1.)
  if abs(D) < 1.e-6: # Identity.
    return x
  else:
    y = np.empty_like(x)
    if abs(B) < 1.e-6:
      qs = np.exp(-D*(SYP-SPP))
      q0 = qs-D*SPP*np.exp(-D*(SYP-SPP))
      qh = 2.-np.exp(-D*(HPP-SYP))
      q1 = qh+D*(1.-HPP)*np.exp(-D*(HPP-SYP))
      q  = 1./(q1-q0)
      # Coefficient for x < SPP.
      b1 = D*np.exp(-D*(SYP-SPP))*q
      # Coefficients for SPP <= x < SYP.
      a2 = -q0*q
      b2 = q
      c2 = -D*SYP
      d2 = D
      # Coefficients for SYP <= x < HPP.
      a3 = (2.-q0)*q
      b3 = -q
      c3 = D*SYP
      d3 = -D
      # Coefficients for x >= HPP.
      a4 = (qh-q0-D*HPP*np.exp(-D*(HPP-SYP)))*q
      b4 = D*np.exp(-D*(HPP-SYP))*q
      # GHS transformation.
      if not inverse:
        mask = (x <  SPP)
        y[mask] =    b1*x[mask]
        mask = (x >= SPP) & (x < SYP)
        y[mask] = a2+b2*np.exp(c2+d2*x[mask])
        mask = (x >= SYP) & (x < HPP)
        y[mask] = a3+b3*np.exp(c3+d3*x[mask])
        mask = (x >= HPP)
        y[mask] = a4+b4*x[mask]
      else:
        SPT =    b1*SPP
        SYT = a2+b2*np.exp(c2+d2*SYP)
        HPT = a4+b4*HPP
        mask = (x <  SPT)
        y[mask] = x[mask]/b1
        mask = (x >= SPT) & (x < SYT)
        y[mask] = (np.log((x[mask]-a2)/b2)-c2)/d2
        mask = (x >= SYT) & (x < HPT)
        y[mask] = (np.log((x[mask]-a3)/b3)-c3)/d3
        mask = (x >= HPT)
        y[mask] = (x[mask]-a4)/b4
    elif abs(B+1.) < 1.e-6:
      qs = -np.log(1.+D*(SYP-SPP))
      q0 = qs-D*SPP/(1.+D*(SYP-SPP))
      qh = np.log(1.+D*(HPP-SYP))
      q1 = qh+D*(1.-HPP)/(1.+D*(HPP-SYP))
      q  = 1./(q1-q0)
      # Coefficient for x < SPP.
      b1 = D/(1.+D*(SYP-SPP))*q
      # Coefficients for SPP <= x < SYP.
      a2 = -q0*q
      b2 = -q
      c2 = 1.+D*SYP
      d2 = -D
      # Coefficients for SYP <= x < HPP.
      a3 = -q0*q
      b3 = q
      c3 = 1.-D*SYP
      d3 = D
      # Coefficients for x >= HPP.
      a4 = (qh-q0-D*HPP/(1.+D*(HPP-SYP)))*q
      b4 = q*D/(1.+D*(HPP-SYP))
      # GHS transformation.
      if not inverse:
        mask = (x <  SPP)
        y[mask] =    b1*x[mask]
        mask = (x >= SPP) & (x < SYP)
        y[mask] = a2+b2*np.log(c2+d2*x[mask])
        mask = (x >= SYP) & (x < HPP)
        y[mask] = a3+b3*np.log(c3+d3*x[mask])
        mask = (x >= HPP)
        y[mask] = a4+b4*x[mask]
      else:
        SPT =    b1*SPP
        SYT = a2+b2*np.log(c2+d2*SYP)
        HPT = a4+b4*HPP
        mask = (x <  SPT)
        y[mask] = x[mask]/b1
        mask = (x >= SPT) & (x < SYT)
        y[mask] = (np.exp((x[mask]-a2)/b2)-c2)/d2
        mask = (x >= SYT) & (x < HPT)
        y[mask] = (np.exp((x[mask]-a3)/b3)-c3)/d3
        mask = (x >= HPT)
        y[mask] = (x[mask]-a4)/b4
    else:
      if B < 0.:
        B  = -B
        qs = (1.-(1.+D*B*(SYP-SPP))**((B-1.)/B))/(B-1.)
        q0 = qs-D*SPP*(1.+D*B*(SYP-SPP))**(-1./B)
        qh = ((1.+D*B*(HPP-SYP))**((B-1.)/B)-1.)/(B-1.)
        q1 = qh+D*(1.-HPP)*(1.+D*B*(HPP-SYP))**(-1./B)
        q  = 1./(q1-q0)
        # Coefficient for x < SPP.
        b1 = D*(1.+D*B*(SYP-SPP))**(-1./B)*q
        # Coefficients for SPP <= x < SYP.
        a2 = (1./(B-1.)-q0)*q
        b2 = -q/(B-1.)
        c2 = 1.+D*B*SYP
        d2 = -D*B
        e2 = (B-1.)/B
        # Coefficients for SYP <= x < HPP.
        a3 = (-1./(B-1.)-q0)*q
        b3 = q/(B-1.)
        c3 = 1.-D*B*SYP
        d3 = D*B
        e3 = (B-1.)/B
        # Coefficients for x >= HPP.
        a4 = (qh-q0-D*HPP*(1.+D*B*(HPP-SYP))**(-1./B))*q
        b4 = D*(1.+D*B*(HPP-SYP))**(-1./B)*q
      else:
        qs = (1.+D*B*(SYP-SPP))**(-1./B)
        q0 = qs-D*SPP*(1.+D*B*(SYP-SPP))**(-(1.+B)/B)
        qh = 2.-(1.+D*B*(HPP-SYP))**(-1./B)
        q1 = qh+D*(1.-HPP)*(1.+D*B*(HPP-SYP))**(-(1.+B)/B)
        q  = 1./(q1-q0)
        # Coefficient for x < SPP.
        b1 = D*(1.+D*B*(SYP-SPP))**(-(1.+B)/B)*q
        # Coefficients for SPP <= x < SYP.
        a2 = -q0*q
        b2 = q
        c2 = 1.+D*B*SYP
        d2 = -D*B
        e2 = -1./B
        # Coefficients for SYP <= x < HPP.
        a3 = (2.-q0)*q
        b3 = -q
        c3 = 1.-D*B*SYP
        d3 = D*B
        e3 = -1./B
        # Coefficients for x >= HPP.
        a4 = (qh-q0-D*HPP*(1.+D*B*(HPP-SYP))**(-(B+1.)/B))*q
        b4 = D*(1.+D*B*(HPP-SYP))**(-(B+1.)/B)*q
      # GHS transformation.
      if not inverse:
        mask = (x <  SPP)
        y[mask] =    b1*x[mask]
        mask = (x >= SPP) & (x < SYP)
        y[mask] = a2+b2*(c2+d2*x[mask])**e2
        mask = (x >= SYP) & (x < HPP)
        y[mask] = a3+b3*(c3+d3*x[mask])**e3
        mask = (x >= HPP)
        y[mask] = a4+b4*x[mask]
      else:
        SPT =    b1*SPP
        SYT = a2+b2*(c2+d2*SYP)**e2
        HPT = a4+b4*HPP
        mask = (x <  SPT)
        y[mask] = x[mask]/b1
        mask = (x >= SPT) & (x < SYT)
        y[mask] = (((x[mask]-a2)/b2)**(1./e2)-c2)/d2
        mask = (x >= SYT) & (x < HPT)
        y[mask] = (((x[mask]-a3)/b3)**(1./e3)-c3)/d3
        mask = (x >= HPT)
        y[mask] = (x[mask]-a4)/b4
    return y



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated (+).

"""Non-local means filter tool."""

from ..gtk.customwidgets import HBox, VBox, CheckButton, HScaleSpinButton
from ..toolmanager import BaseToolWindow
from skimage.restoration import estimate_sigma, denoise_nl_means

class NonLocalMeansFilterTool(BaseToolWindow):
  """Non-local means filter tool class."""

  _action_ = "Applying non-local means filter..."

  _help_ = """Non-local means filter for denoising:

    OUT(r) \u221d \u03a3r' IMG(r')f(r, r')

where:

    f(r, r') = exp[-(M(r)-M(r'))²/h²]

and M(r) is an average of the pixels in a patch around r. The filter is controlled by:
  \u2022 The size of the (square) patch used to compute M(r). The pixels within the patch are uniformly averaged if "fast algorithm" is ticked, weighted by a gaussian if not (better yet slower).
  \u2022 The maximal distance |r-r'| between the patches.
  \u2022 The cut-off h in gray levels (the filter is applied to the red, blue and green channels independently).
The standard deviation sigma of the noise may be provided and subtracted out when computing f(r, r'). This can lead to a modest improvement in image quality.
The non-local means filter can restore textures that would be blurred by other denoising algorithms."""

  _onthefly_ = False # This transformation can not be applied on the fly.

  def open(self, image):
    """Open tool window for image 'image'."""
    if not super().open(image, "Non-local means filter"): return False
    self.sigma = estimate_sigma(self.reference.rgb, channel_axis = 0, average_sigmas = True)
    wbox = VBox()
    self.window.add(wbox)
    self.widgets.psizescale = HScaleSpinButton(7., 1., 25., 1., digits = 0, length = 480)
    wbox.pack(self.widgets.psizescale.layout2("Patch size (pixels):"))
    self.widgets.pdistscale = HScaleSpinButton(11., 1., 50., 1., digits = 0, length = 480)
    wbox.pack(self.widgets.pdistscale.layout2("Patch distance (pixels):"))
    self.widgets.cutoffscale = HScaleSpinButton(.01, 0., .1, .0001, digits = 4, length = 480)
    wbox.pack(self.widgets.cutoffscale.layout2("Cut-off (gray levels):"))
    hbox = HBox()
    wbox.pack(hbox)
    self.widgets.sigmabutton = CheckButton(label = f"Use estimated noise rms = {self.sigma:.5e}")
    self.widgets.sigmabutton.set_active(True)
    hbox.pack(self.widgets.sigmabutton)
    self.widgets.fastbutton = CheckButton(label = "Use fast algorithm")
    self.widgets.fastbutton.set_active(True)
    hbox.pack(self.widgets.fastbutton)
    wbox.pack(self.tool_control_buttons())
    self.start(identity = False)
    return True

  def get_params(self):
    """Return tool parameters."""
    return int(round(self.widgets.psizescale.get_value())), int(round(self.widgets.pdistscale.get_value())), self.widgets.cutoffscale.get_value(), \
           self.widgets.sigmabutton.get_active(), self.widgets.fastbutton.get_active()

  def set_params(self, params):
    """Set tool parameters 'params'."""
    psize, pdist, cutoff, sigma, fast = params
    self.widgets.psizescale.set_value(psize)
    self.widgets.pdistscale.set_value(pdist)
    self.widgets.cutoffscale.set_value(cutoff)
    self.widgets.sigmabutton.set_active(sigma)
    self.widgets.fastbutton.set_active(fast)

  def run(self, params):
    """Run tool for parameters 'params'."""
    psize, pdist, cutoff, sigma, fast = params
    if psize <= 0 or pdist <= 0 or cutoff <= 0.: return params, False
    self.image.rgb = denoise_nl_means(self.reference.rgb, channel_axis = 0, patch_size = psize, patch_distance = pdist, h = cutoff, \
                                      sigma = self.sigma if sigma else 0., fast_mode = fast)
    return params, True

  def operation(self, params):
    """Return tool operation string for parameters 'params'."""
    psize, pdist, cutoff, sigma, fast = params
    optstring = ""
    if sigma: optstring += f", sigma = {self.sigma:.5e}"
    if fast: optstring += ", fast"
    return f"NonLocalMeansFilter(patch size = {psize} pixels, patch distance = {pdist} pixels, cutoff = {cutoff:.4f}{optstring})"



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated (+).

"""Pixel math tool."""

from ..gtk.utils import markup_escape_text
from ..gtk.customwidgets import VBox, FramedVBox, ScrolledBox, Entry, TextView
from ..toolmanager import BaseToolWindow
from ..misc.imagechooser import ImageChooser
from ...imageprocessing.utils import is_valid_rgb_image
from ...imageprocessing.pixelmath import PixelMath
import numpy as np
from collections import OrderedDict as OD

class PixelMathTool(BaseToolWindow):
  """Pixel math window class."""

  _action_ = "Doing pixel math..."

  _onthefly_ = False # This transformation can not be applied on the fly.

  _referencetab_ = False # Don't show the reference image tab.

  _HELP_ = """<b>Instructions</b>:
Use python syntax. Reference image #i of the above list as 'IMGi'. Module numpy is imported as np.
<b>Commands</b>:
  \u2022 value(IMG1, midtone = .5): HSV value of 'IMG1', with midtone correction 'midtone'.
  \u2022 luma(IMG1, midtone = .5): luma of 'IMG1', with midtone correction 'midtone'.
  \u2022 luminance(IMG1, midtone = .5): luminance of 'IMG1', with midtone correction 'midtone'
        (here, luminance is defined as a gray scale image with the same luminance/lightness as IMG1).
  \u2022 lightness(IMG1, midtone = .5): CIE lightness of 'IMG1', with midtone correction 'midtone' (within [0, 100]).
  \u2022 scale(IMG1, source, target): Scale 'IMG1' by the ratio 'target'/'source' (wherever source \u2260 0).
  \u2022 blend(IMG1, IMG2, mix): Returns (1-mix)*IMG1+mix*IMG2. 'mix' can be an image or a scalar.
  \u2022 mts(IMG1, midtone): Apply midtone stretch function to IMG1.
  \u2022 ghs(IMG1, lnD1, B, SYP, SPP = 0, HPP = 1): Apply generalized hyperbolic stretch function to IMG1.
<b>Example</b>:
  \u2022 HDR composition between "short" exposure image 'IMG1', "medium" exposure image 'IMG2', and "long" exposure image 'IMG3':
        blend(IMG3, blend(IMG2, IMG1, luminance(IMG1)), luminance(IMG3))
  \u2022 Recombination of a starless 'IMG1' and a star mask 'IMG2':
        1-(1-IMG1)*(1-IMG2)

<b>Log</b>:
"""

  def open(self, image):
    """Open tool window for image 'image'."""
    if not super().open(image, "Pixel math"): return False
    self.app.mainwindow.set_images(OD(Image = self.image, Selection = self.image), reference = "Selection") # Add the selection as the reference image.
    wbox = VBox()
    self.window.add(wbox)
    wbox.pack("List of available images:")
    self.widgets.chooser = ImageChooser(self.app, self.window, wbox, tabkey = "Selection", last = True)
    self.widgets.chooser.set_selected_row(-1)
    frame, vbox = FramedVBox()
    wbox.pack(frame, expand = True, fill = True)
    self.widgets.scrolled = ScrolledBox(800, 200)
    vbox.pack(self.widgets.scrolled, expand = True, fill = True)
    self.widgets.textview = TextView(wrap = False)
    self.widgets.textview.append_markup(self._HELP_)
    self.widgets.scrolled.add(self.widgets.textview)
    self.widgets.commandentry = Entry()
    self.widgets.commandentry.connect("activate", lambda entry: self.apply())
    wbox.pack(self.widgets.commandentry.hbox(prepend = "IMG = "))
    wbox.pack(self.tool_control_buttons())
    self.start(identity = True)
    return True

  def get_params(self):
    """Return tool parameters."""
    return self.widgets.commandentry.get_text()

  def set_params(self, params):
    """Set tool parameters 'params'."""
    command = params
    self.widgets.commandentry.set_text(command)

  def run(self, params):
    """Run tool for parameters 'params'."""
    command = params.strip()
    if command == "": return params, False
    try:
      pm = PixelMath(self.widgets.chooser.get_images_list())
      output = pm.run(command)
    except Exception as err:
      self.queue_gui_mainloop(self.append_message, markup_escape_text(str(err)+"."), "red")
      return "", False
    if not is_valid_rgb_image(output):
      self.queue_gui_mainloop(self.append_message, "The command did not return a valid image.", "red")
      return "", False
    self.image.set_image(output)
    self.queue_gui_mainloop(self.append_message, "Done.", "green")
    return params, True

  def operation(self, params):
    """Return tool operation string for parameters 'params'."""
    command = params.strip()
    if command == "": return None
    files = ""
    separator = ""
    for image in self.widgets.chooser.get_images_list():
      filename = image.meta.get("imchooser.file", None)
      if filename:
        row = image.meta.get("imchooser.row")
        files += f"{separator}#{row}: '{filename}'"
        separator = ", "
    if files: files = ", {"+files+"}"
    return f"PixelMath('{command}'{files})"

  # Log messages.

  def append_message(self, message, color = "black"):
    """Append message 'message' with color 'color' to self.widgets.textview."""
    self.widgets.textview.append_markup(f"<span foreground='{color}'>{message}</span>\n")
    vadj = self.widgets.scrolled.get_vadjustment() # Display the end of the text buffer.
    vadj.set_value(vadj.get_upper()-vadj.get_page_size())



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated (+).

"""Switch image tool."""

from ..gtk.customwidgets import VBox
from ..toolmanager import BaseToolWindow
from ..misc.imagechooser import ImageChooser

class SwitchTool(BaseToolWindow):
  """Switch tool window class."""

  _action_ = "Switching to an other image..."

  _onthefly_ = False # This tool is actually applied on the fly, but uses its own signals & callbacks to track changes.

  _referencetab_ = False # Don't show the reference image tab.

  def open(self, image):
    """Open tool window for image 'image'."""
    if not super().open(image, "Switch image"): return False
    wbox = VBox()
    self.window.add(wbox)
    wbox.pack("Choose image to switch to:")
    self.widgets.chooser = ImageChooser(self.app, self.window, wbox, tabkey = None, callback = lambda row, image: self.apply())
    wbox.pack(self.tool_control_buttons(model = "onthefly", reset = False))
    self.start(identity = True)
    return True

  def get_params(self):
    """Return tool parameters."""
    return self.widgets.chooser.get_selected_row()

  def set_params(self, params):
    """Set tool parameters 'params'."""
    row = params
    self.widgets.chooser.set_selected_row(row)

  def run(self, params):
    """Run tool for parameters 'params'."""
    row = params
    if row < 0: return params, False
    self.image.copy_image_from(self.widgets.chooser.get_image(row))
    #self.frame = self.image.get_frame()
    return params, True

  def operation(self, params):
    """Return tool operation string for parameters 'params'."""
    row = params
    return f"SwitchTo({self.widgets.chooser.get_image_tag(row)})" if row >= 0 else None



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17

"""Signals management."""

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

class Signals:
  """Add extended signal management to a given widget class
     See gtk/customwidgets for examples."""

  def __init__(self):
    """Initialize class."""
    self._signals = {}

  def connect(self, signames, *args, **kwargs):
    """Connect signals 'signames' to the callback defined by (args, kwargs)."""
    if not isinstance(signames, tuple): signames = (signames, )
    for signame in signames:
      self._signals[signame] = super().connect(signame, *args, **kwargs)

  def disconnect(self, signames):
    """Disconnect signals 'signames'."""
    if not isinstance(signames, tuple): signames = (signames, )
    for signame in signames:
      super().disconnect(self._signals[signame])
      del self._signals[signame]

  def block(self, signames):
    """Block signals 'signames'."""
    if not isinstance(signames, tuple): signames = (signames, )
    for signame in signames:
      self.handler_block(self._signals[signame])

  def unblock(self, signames):
    """Unblock signals 'signames'."""
    if not isinstance(signames, tuple): signames = (signames, )
    for signame in signames:
      self.handler_unblock(self._signals[signame])

  def block_all_signals(self):
    """Block all signals."""
    for signal in self._signals.values():
      self.handler_block(signal)

  def unblock_all_signals(self):
    """Unblock all signals."""
    for signal in self._signals.values():
      self.handler_unblock(signal)



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17

"""Custom Gtk file choosers."""

import os
import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GdkPixbuf

def ImageFileChooserDialog(window, action, path = None, preview = False, title = None, extra_widget = None):
  """Open file chooser dialog for an image, from window 'window' and for
     action 'action' (either Gtk.FileChooserAction.OPEN to open an image
     or Gtk.FileChooserAction.SAVE to save an image). Start with directory
     and file name 'path' (default if None), and preview selected image
     if 'preview' is True. If not None, 'title' overrides the default title
     of the dialog, and 'extra_widget' is a widget to provide, for example,
     additional options.
     Return chosen file name or None if cancelled."""

  def update_preview(dialog):
    """Update preview."""
    filename = dialog.get_preview_filename()
    try:
      pixbuf = GdkPixbuf.Pixbuf.new_from_file(filename)
    except:
      dialog.set_preview_widget_active(False)
    else:
      maxwidth, maxheight = 360, 720 # Scale image.
      width, height = pixbuf.get_width(), pixbuf.get_height()
      scale = min(maxwidth/width, maxheight/height)
      if scale < 1:
        width, height = int(width*scale), int(height*scale)
        pixbuf = pixbuf.scale_simple(width, height, GdkPixbuf.InterpType.BILINEAR)
        preview_image.set_from_pixbuf(pixbuf)
        dialog.set_preview_widget_active(True)

  if action == Gtk.FileChooserAction.OPEN:
    title_ = "Open image"
    button = Gtk.STOCK_OPEN
  elif action == Gtk.FileChooserAction.SAVE:
    title_ = "Save image as"
    button = Gtk.STOCK_SAVE
  else:
    raise ValueError("action must be Gtk.FileChooserAction.OPEN or Gtk.FileChooserAction.SAVE.")
  if title is None: title = title_
  dialog = Gtk.FileChooserDialog(title = title, transient_for = window, action = action)
  dialog.add_buttons(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                     button, Gtk.ResponseType.OK)
  dialog.set_do_overwrite_confirmation(True)
  if extra_widget is not None: dialog.set_extra_widget(extra_widget)
  if preview and action == Gtk.FileChooserAction.OPEN:
    preview_image = Gtk.Image()
    dialog.set_preview_widget(preview_image)
    dialog.connect("update-preview", update_preview)
  if path is not None:
    dialog.set_filename(path)
    if action == Gtk.FileChooserAction.SAVE: dialog.set_current_name(os.path.basename(path))
  response = dialog.run()
  filename = dialog.get_filename()
  dialog.destroy()
  return filename if response == Gtk.ResponseType.OK else None



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated.

"""Base tool window class."""

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GObject
from .gtk.customwidgets import Label, HBox, VBox, HButtonBox, Button
from .gtk.keyboard import decode_key
from .base import BaseWindow, Container
import threading
from collections import OrderedDict as OD

class BaseToolWindow(BaseWindow):
  """Base tool window class."""

  _action_ = None # Message printed when the tool is applied.

  _help_ = None # Help message.

  _onthefly_ = True # True if the transformations can be applied on the fly.

  _referencetab_ = True # Show the reference image tab.

  def __init__(self, app, polltime = -1):
    """Bind window with application 'app'.
       If polltime > 0, run the tool on the fly by polling for
       tool parameters change every 'polltime' ms.
       If polltime <= 0, run the tool on demand when the
       user clicks the "Apply" button."""
    super().__init__(app)
    self.polltime = polltime if self._onthefly_ else -1
    self.onthefly = (self.polltime > 0)

  def open(self, image, title):
    """Open tool window with title 'title' for image 'image'.
       Return True if successful, False otherwise."""
    if self.opened: return False
    if self._action_ is not None: print(self._action_)
    self.opened = True
    self.image = image.clone()
    self.image.meta["params"] = None
    self.image.meta["description"] = "[No transformations]"
    self.image.stats = None # Image statistics.
    self.reference = image.ref()
    self.reference.meta["description"] = "Reference image"
    self.reference.stats = None # Reference image statistics.
    self.window = Gtk.Window(title = title, border_width = 16)
    self.window.connect("delete-event", self.quit)
    self.window.connect("key-press-event", self.key_press)
    self.widgets = Container()
    if self._referencetab_:
      self.app.mainwindow.set_images(OD(Image = self.image, Reference = self.reference), reference = "Reference")
    else:
      self.app.mainwindow.set_images(OD(Image = self.image))
    self.app.mainwindow.set_copy_paste_callbacks(self.copy, self.paste)
    self.polltimer = None # Polling/update threads data.
    self.lock = threading.Lock()
    self.thread = threading.Thread(target = None)
    self.toolparams = None # Tool parameters of the last transformation.
    self.transformed = False # True if the image has been transformed.
    self.defaultparams = None # Default tool parameters.
    self.defaultparams_are_identity = True # True if default tool parameters are the identity operation.
    self.frame = None # New frame if modified by the tool.
    return True

  # Start tool.

  def start(self, identity = True):
    """Start tool.
       Set the present tool parameters (drawn from self.get_params()) as default parameters, update the
       main window tabs, show the tool window and start polling for tool parameters change.
       If 'identity' is True, the default parameters are the identity operation (no image transformation)."""
    self.defaultparams = self.get_params()
    self.toolparams = self.get_params()
    self.defaultparams_are_identity = identity
    if self.onthefly and not identity: self.apply(cancellable = False)
    self.window.show_all()
    self.start_polling()

  # Finalize tool.

  def finalize(self, image, operation, frame = None):
    """Finalize tool.
       Close window and return image 'image' (if not None), operation 'operation', and frame 'frame' to the application."""
    self.app.mainwindow.set_copy_paste_callbacks(None, None) # Disconnect Ctrl-C/Ctrl-V callbacks.
    self.app.mainwindow.set_rgb_luma_callback(None) # Disconnect RGB luma callback (if any).
    self.app.mainwindow.set_guide_lines(None) # Remove guide lines.
    self.window.destroy()
    self.opened = False
    if image is not None:
      image.meta.pop("params", None) # Clean-up the image meta-data.
      self.app.finalize_tool(image, operation, frame)
    del self.widgets
    del self.image
    del self.reference
    self.cleanup() # Additional cleanup.

  def destroy(self):
    """Destroy tool (without returning any image and operation to the application)."""
    if not self.opened: return
    self.stop_polling() # Stop polling.
    if self.thread.is_alive(): self.thread.join() # Wait for the current thread to finish.
    self.finalize(None, None)

  def quit(self, *args, **kwargs):
    """Quit tool (return reference image and operation = None to the application)."""
    if not self.opened: return
    self.stop_polling() # Stop polling.
    if self.thread.is_alive(): self.thread.join() # Wait for the current thread to finish.
    self.finalize(self.reference, None)

  def close(self, *args, **kwargs):
    """Close tool (return current image, operation and frame to the application)."""
    if not self.opened: return
    polling = self.stop_polling() # Stop polling.
    if self.thread.is_alive(): self.thread.join() # Wait for the current thread to finish.
    if polling:
      params = self.get_params()
      if params != self.toolparams: # Make sure that the last changes have been applied.
        self.start_run_thread(params)
        if self.thread.is_alive(): self.thread.join() # Shall be alive, actually.
    self.finalize(self.image, self.image.meta["description"] if self.transformed else None, self.frame)

  def cleanup(self):
    """Free memory on exit.
       Must be defined (if needed) in each subclass."""
    return

  # Tool control buttons.

  def tool_control_buttons(self, model = None, reset = True):
    """Return a HBox with tool control buttons.
       If None, 'model' is set to "ondemand" if self.onthefly is False, and to "onthefly" if self.onthefly is True.
       If 'model' is "ondemand", the transformations are applied on demand and the control buttons are
         Apply, Cancel, Reset and Close
       connected to the self.apply, self.cancel, self.reset and self.close methods.
       If 'model' is "onthefly", the transformations are applied on the fly and the control buttons are
         OK, Reset, and Cancel
       connected to the self.close, self.cancel and self.quit methods.
       If 'model' is "applyonce", the tool closes once the transformation is applied on demand and the control buttons are
         Apply, Reset, and Cancel
       connected to the self.apply_and_close, self.reset and self.quit methods.
       The Reset button is not displayed if 'reset' is False."""
    if model is None:
      model = "onthefly" if self.onthefly else "ondemand"
    hbox = HButtonBox()
    if model == "ondemand":
      self.widgets.applybutton = Button(label = "Apply") # Apply transformation on demand.
      self.widgets.applybutton.connect("clicked", self.apply)
      hbox.pack(self.widgets.applybutton)
      self.widgets.cancelbutton = Button(label = "Cancel") # Cancel transformation (restore the reference image).
      self.widgets.cancelbutton.connect("clicked", self.cancel)
      self.widgets.cancelbutton.set_sensitive(False)
      hbox.pack(self.widgets.cancelbutton)
      self.widgets.resetbutton = Button(label = "Reset") # Reset parameters to the last applied transformation.
      self.widgets.resetbutton.connect("clicked", self.reset)
      if reset: hbox.pack(self.widgets.resetbutton)
      self.widgets.closebutton = Button(label = "Close") # Close tool and return the transformed image to the application.
      self.widgets.closebutton.connect("clicked", self.close)
      hbox.pack(self.widgets.closebutton)
    elif model == "onthefly":
      self.widgets.applybutton = None
      self.widgets.closebutton = Button(label = "OK") # Close tool and return the transformed image to the application.
      self.widgets.closebutton.connect("clicked", self.close)
      hbox.pack(self.widgets.closebutton)
      self.widgets.cancelbutton = Button(label = "Reset") # Cancel transformation (restore the reference image).
      self.widgets.cancelbutton.connect("clicked", self.cancel)
      self.widgets.cancelbutton.set_sensitive(False)
      if reset: hbox.pack(self.widgets.cancelbutton)
      self.widgets.quitbutton = Button(label = "Cancel") # Cancel transformation and close tool (return the reference image to the application).
      self.widgets.quitbutton.connect("clicked", self.quit)
      hbox.pack(self.widgets.quitbutton)
    elif model == "applyonce":
      self.widgets.applybutton = Button(label = "Apply") # Apply transformation on demand and return the transformed image to the application.
      self.widgets.applybutton.connect("clicked", self.apply_and_close)
      hbox.pack(self.widgets.applybutton)
      self.widgets.resetbutton = Button(label = "Reset") # Reset parameters.
      self.widgets.resetbutton.connect("clicked", self.reset)
      if reset: hbox.pack(self.widgets.resetbutton)
      self.widgets.quitbutton = Button(label = "Cancel") # Close tool.
      self.widgets.quitbutton.connect("clicked", self.quit)
      hbox.pack(self.widgets.quitbutton)
    else:
      raise ValueError("Model must be 'onthefly', 'ondemand', or 'applyonce'.")
    return self.add_help_tooltip(hbox)

  def add_help_tooltip(self, box):
    """Add a "[?]" label with a help tooltip at the right of box 'box' (if self._help_ is not None).
       Return a new HBox() embedding box and label."""
    if self._help_ is None: return box
    hbox = HBox()
    hbox.pack(box, expand = True, fill = True)
    label = Label("[?]")
    label.set_tooltip_text(self._help_)
    hbox.pack(label)
    return hbox

  # Apply tool.

  def get_params(self):
    """Return tool parameters.
       Must be defined in each subclass."""
    return None

  def set_params(self, params):
    """Set tool parameters 'params'.
       Must be defined in each subclass."""
    return

  def run(self, params):
    """Run tool for parameters 'params'.
       Return (toolparams, transformed), where toolparams are the tool parameters
       actually applied (that may differ from params if the latter are, e.g., out
       of range), and transformed is True if self.image has indeed been transformed
       (with respect to self.reference), False otherwise.
       Must be defined in each subclass."""
    print("Doing nothing !...")
    return None, False

  def operation(self, params):
    """Return tool operation string for parameters 'params'.
       Must be defined in each subclass."""
    return None

  #

  def queue_gui_mainloop(self, function, *args):
    """Enqueue a thread-safe call to 'function(*args)' in the GUI mainloop.
       This is a wrapper to GObject.idle_add(function, *args, priority = GObject.PRIORITY_DEFAULT)."""
    GObject.idle_add(function, *args, priority = GObject.PRIORITY_DEFAULT)

  def update_gui(self):
    """Update main and tool windows after tool run."""
    if not self.opened: return
    self.app.mainwindow.update_image("Image", self.image)
    self.app.mainwindow.update_key_label("Image", "Image (*)" if self.transformed else "Image")
    if not self.thread.is_alive():
      self.app.mainwindow.set_idle()
      self.app.mainwindow.unlock_rgb_luma()
      if self.widgets.applybutton is not None: self.widgets.applybutton.set_sensitive(True)
    return False

  def start_run_thread(self, params):
    """Run tool for params 'params' and update main and tool windows in a separate thread to keep the GUI responsive."""
    if self.thread.is_alive(): self.thread.join() # Wait for the current thread to finish.
    self.app.mainwindow.lock_rgb_luma()
    self.app.mainwindow.set_busy()
    self.thread = threading.Thread(target = self.run_thread, args = (params,), daemon = False)
    self.thread.start()

  def run_thread(self, params):
    """Run tool for params 'params' and update main and tool windows."""
    with self.lock: # Make sure no other thread is running concurrently.
      toolparams, self.transformed = self.run(params) # Must be defined in each subclass.
      if not self.transformed: self.image.copy_image_from(self.reference)
      self.image.meta["params"] = toolparams
      self.image.meta["description"] = self.operation(toolparams)
      self.toolparams = params
      self.queue_gui_mainloop(self.update_gui) # Thread-safe.

  def apply(self, *args, **kwargs):
    """Get tool parameters, run tool and update main and tool windows.
       If the keyword argument 'cancellable' is False (default True), this run can not be cancelled,
       so that the "Cancel" button is made insensitive."""
    params = self.get_params()
    if params is None: return # Do nothing is params is None.
    if self.widgets.applybutton is not None: self.widgets.applybutton.set_sensitive(False)
    self.start_run_thread(params)
    cancellable = kwargs.get("cancellable", True)
    if cancellable is not None: self.widgets.cancelbutton.set_sensitive(cancellable)
    #self.app.mainwindow.set_current_tab(0)

  def apply_and_close(self, *args, **kwargs):
    """Get tool parameters and run tool, then close tool and return transformed image."""
    self.apply(cancellable = None)
    self.close()

  def apply_idle(self):
    """Get tool parameters, run tool and update main and tool windows if no other thread is already running.
       Return True if new thread successfully started, False otherwise."""
    if self.thread.is_alive(): return False
    params = self.get_params()
    if params is None: return False # Do nothing is params is None.
    self.start_run_thread(params)
    self.widgets.cancelbutton.set_sensitive(True)
    return True

  # Reset/Cancel tool.

  def reset(self, *args, **kwargs):
    """Reset tool parameters."""
    self.set_params(self.toolparams)

  def cancel(self, *args, **kwargs):
    """Cancel tool."""
    self.stop_polling() # Stop polling while restoring reference image.
    if self.thread.is_alive(): self.thread.join() # Wait for the current thread to finish.
    self.set_params(self.defaultparams)
    if self.onthefly and not self.defaultparams_are_identity:
      self.apply(cancellable = False)
    else:
      self.image.copy_image_from(self.reference)
      self.image.meta["params"] = None
      self.image.meta["description"] = "[No transformations]"
      self.toolparams = self.get_params()
      self.transformed = False
      self.update_gui()
    self.widgets.cancelbutton.set_sensitive(False)
    self.resume_polling() # Resume polling.

  # Polling for tool parameter changes.

  def start_polling(self, lastparams = None):
    """Start polling for tool parameter changes every self.polltime ms.
       At each poll, get the tool parameters from self.get_params(); then
       call self.apply_idle() if the tool parameters are the same *twice* in a row,
       but are different from the self.toolparams registered at the last update.
       'lastparams' is the assumptive outcome of the last poll (defaults to self.toolparams if None);
       set to self.get_params() to expedite call to self.apply_idle() as soon as the first poll.
       Return true if successfully polling, False otherwise (self.polltime < 0)."""
    if self.polltime <= 0: return False # No poll time defined.
    self.pollparams = self.toolparams if lastparams is None else lastparams
    self.polltimer = GObject.timeout_add(self.polltime, self.poll)
    return True

  def poll(self, *args, **kwargs):
    """Poll for tool parameter changes, and call self.apply_idle()
       if the tool parameters are the same *twice* in a row, but are
       different from the self.toolparams registered at the last update."""
    params = self.get_params()
    if params != self.toolparams and params == self.pollparams: self.apply_idle()
    self.pollparams = params
    return True

  def stop_polling(self):
    """Stop polling for tool parameter changes.
       Return True if polling was actually enabled, False otherwise."""
    ispolling = self.polltimer is not None
    if ispolling:
      GObject.source_remove(self.polltimer)
      self.polltimer = None
    return ispolling

  def resume_polling(self):
    """Resume polling for tool parameter changes.
       Return true if successfully polling, False otherwise (self.polltime < 0)."""
    if self.polltimer is not None: return True # Already polling.
    return self.start_polling()

  def reset_polling(self, lastparams = None):
    """Reset polling for tool parameter changes (call stop_polling/start_polling in a row).
       Return true if successfully polling, False otherwise (self.polltime < 0)."""
    if self.polltimer is None: return False
    self.stop_polling()
    return self.start_polling(lastparams)

  def connect_update_request(self, widget, signames):
    """Connect signals 'signames' of widget 'widget' to self.reset_polling(self.get_params()) in
       order to request tool update on next poll. This enhances responsivity to tool parameters changes."""
    widget.connect(signames, lambda *args: self.reset_polling(self.get_params()))

  # Manage key press/release events.

  def key_press(self, widget, event):
    """Callback for key press in the tool window."""
    key = decode_key(event)
    if key.ctrl and not key.alt and key.uname == "TAB":
      self.app.mainwindow.set_current_tab(0)
      self.app.mainwindow.window.present()

  # Ctrl-C/Ctrl-V callbacks.

  def copy(self, key, image):
    """Copy image 'image' with key 'key' in a new tab."""
    if key != "Image": return # Can only copy the transformed image.
    if image.meta["params"] is None: return
    keys = self.app.mainwindow.get_keys()
    ncopies = [key[0:4] == "Copy" for key in keys].count(True)
    if ncopies >= 10: return # Allow 10 copies max.
    ncopies += 1
    clone = image.clone()
    clone.meta["tag"] = f"#{ncopies}"
    clone.meta["deletable"] = True
    self.app.mainwindow.append_image(f"Copy #{ncopies}", clone)

  def paste(self, key, image):
    """Paste the parameters of the image 'image' with key 'key' to the tool."""
    if key[0:4] != "Copy": return # Can only paste the parameters from the copies.
    params = image.meta["params"]
    self.set_params(params)
    self.reset_polling(params)



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated.

"""Tool utils."""

import numpy as np
import matplotlib as mpl
import matplotlib.colors as colors
import matplotlib.ticker as ticker

def histogram_bins(stats, colordepth = 8):
  """Return the number of histogram bins for an image with (luma) stats 'stats' (see imageprocessing.Image.statistics)
     and color depth 'colordepth' (bits/color)."""
  iqr = stats.percentiles[2]-stats.percentiles[0] if stats.percentiles is not None else .25
  nbins = int(stats.npixels**(1./3.)/(2.*iqr))
  nbinsmin = 128
  nbinsmax = min(2**(colordepth-1), 8192)
  return min(max(nbins, nbinsmin), nbinsmax)

def plot_histograms(ax, edges, counts, colors,
                    title = None, xlabel = "Level", ylabel = "Count (a.u.)", ylogscale = False):
  """Plot histograms with bin edges 'edges(nbins)' and bin counts 'counts(nc, nbins)' in axes 'ax',
     where nc is the number of histogram channels and nbins the number of histogram bins.
     Use color 'colors(nc)' for the histogram lines (not displayed if None).
     Set title 'title', x label 'xlabel' and y label 'ylabel' (if not None).
     Use log scale on y axis if 'ylogscale' is True.
     Return a list of nc matplotlib.lines.Line2D histogram lines."""
  centers = (edges[:-1]+edges[1:])/2.
  imin = np.argmin(abs(centers-0.))
  imax = np.argmin(abs(centers-1.))
  cmax = counts[:, imin+1:imax].max()
  rcounts = counts/cmax
  ax.clear()
  histlines = []
  for ic in range(counts.shape[0]):
    histlines.append(ax.plot(centers, rcounts[ic], "-", color = colors[ic])[0] if colors[ic] is not None else None)
  xmin = min(0., centers[ 0])
  xmax = max(1., centers[-1])
  ax.set_xlim(xmin, xmax)
  ax.xaxis.set_minor_locator(ticker.AutoMinorLocator(5))
  if xlabel is not None: ax.set_xlabel(xlabel)
  if ylogscale:
    ax.set_yscale("log")
    #ax.set_ylim(1./cmax, 1.)
    ax.set_ylim(rcounts[counts > 0.].min(), 1.)
  else:
    ax.set_yscale("linear")
    ax.set_ylim(0., 1.)
    ax.yaxis.set_minor_locator(ticker.AutoMinorLocator(5))
  if ylabel is not None: ax.set_ylabel(ylabel)
  ax.axvspan(xmin-1., 0., color = "gray", alpha = .25)
  ax.axvspan(1., xmax+1., color = "gray", alpha = .25)
  if title is not None: ax.set_title(title, weight = "bold")
  return histlines

def update_histograms(ax, histlines, edges, counts, ylogscale = False):
  """Update histogram lines 'histlines(nc)' in axes 'ax' with bin edges 'edges(nbins)' and bin
     counts 'counts(nc, nbins)', where nc is the number of histogram channels and nbins the
     number of histogram bins. Use log scale on y axis if 'ylogscale' is True."""
  centers = (edges[:-1]+edges[1:])/2.
  imin = np.argmin(abs(centers-0.))
  imax = np.argmin(abs(centers-1.))
  cmax = counts[:, imin+1:imax].max()
  rcounts = counts/cmax
  for ic in range(len(histlines)):
    if histlines[ic] is not None:
      histlines[ic].set_xdata(centers)
      histlines[ic].set_ydata(rcounts[ic])
  if ylogscale:
    ax.set_yscale("log")
    #ax.set_ylim(1./cmax, 1.)
    ax.set_ylim(rcounts[counts > 0.].min(), 1.)
  else:
    ax.set_yscale("linear")
    ax.set_ylim(0., 1.)
    ax.yaxis.set_minor_locator(ticker.AutoMinorLocator(5))

def highlight_histogram(histlines, idx, lw = mpl.rcParams["lines.linewidth"]):
  """Highlight histogram line 'histlines[idx]' by making it twice thicker and bringing it to front. 'lw' is the default linewidth."""
  for ic in range(len(histlines)):
    line = histlines[ic]
    if line is None: continue
    if ic == idx:
      line.set_zorder(3)
      line.set_linewidth(lw*1.4142)
    else:
      line.set_zorder(2)
      line.set_linewidth(lw/1.4142)

def transform_histogram(edges, counts, fedges, newedges):
  """Transform histogram defined by bin edges 'edges' and bin counts 'counts'
     by applying a function 'fedges' = f(edges) to the edges then resampling over
     new edges 'newedges'. Return the new counts."""
  nedges = len(edges)
  csum = np.empty(nedges+1, dtype = counts.dtype)
  csum[0] = 0
  np.cumsum(counts, out = csum[1:])
  fcsum = np.interp(newedges, fedges, csum)
  return np.diff(fcsum)

def stats_string(stats):
  """Return string for the channel statistics 'stats' (see imageprocessing.Image.statistics)."""
  median = f"{stats.median:.5f}" if stats.median is not None else "None"
  return f"{stats.name} : min = {stats.minimum:.5f}, max = {stats.maximum:.5f}, med = {median}, {stats.zerocount} ({100.*stats.zerocount/stats.npixels:.2f}%) zeros, {stats.outcount} ({100.*stats.outcount/stats.npixels:.2f}%) out-of-range"

def plot_hsv_wheel(ax):
  """Plot HSV wheel in ploar axes 'ax'."""
  rho = np.linspace(0., 1., 100)
  phi = np.linspace(0., 2.*np.pi, 100)
  RHO, PHI = np.meshgrid(rho, phi)
  h = np.ravel(PHI/(2.*np.pi))
  s = np.ravel(RHO)
  v = np.ones_like(s)
  hsv = np.column_stack((h, s, v))
  rgb = colors.set_hsv_image(hsv)
  ax.scatter(PHI, RHO, c = rgb, zorder = -3)
  ax.set_ylim(0., 1.)



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated (+).

"""Add Unistellar frame from an other image."""

import os
import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
from ..gtk.customwidgets import Label, VBox, FramedHBox, Grid, Button, HoldButton, CheckButton, SpinButton
from ..gtk.filechoosers import ImageFileChooserDialog
from ..base import ErrorDialog
from ..toolmanager import BaseToolWindow
from ...imageprocessing import imageprocessing
import matplotlib.pyplot as plt
import numpy as np

class AddUnistellarFrame(BaseToolWindow):
  """Add Unistellar frame tool class."""

  _action_ = "Adding Unistellar frame..."

  _help_ = """Add the Unistellar frame of an other image."""

  delay = 333 # Long press delay for "HoldButton" (ms).
  maxfade = .05 # Maximum fade.

  def open(self, image):
    """Open tool window for image 'image'."""
    if not super().open(image, "Add frame"): return False
    filename = ImageFileChooserDialog(self.app.mainwindow.window, Gtk.FileChooserAction.OPEN, path = self.app.get_filename(), preview = True, title = "Open framed image")
    if filename is None:
      self.destroy()
      return False
    try:
      image = self.app.ImageClass()
      image.load(filename)
    except Exception as err:
      ErrorDialog(self.window, str(err))
      self.destroy()
      return False
    framed = image.is_valid()
    if framed: framed = image.check_frame()
    if not framed:
      ErrorDialog(self.window, "This image is not valid or has no frame.")
      self.destroy()
      return False
    print(f"""Image has a frame type '{framed["type"]}'.""")
    self.basename = os.path.basename(filename)
    self.frame = image.get_frame()
    self.fradius = framed["radius"]
    self.fmargin = framed["margin"]
    self.fwidth, self.fheight = self.frame.size()
    self.xcenter = 0
    self.ycenter = 0
    self.currentmove = None
    self.currentscale = None
    self.currentmarginfade = None
    wbox = VBox()
    self.window.add(wbox)
    self.widgets.marginspin = SpinButton(self.fmargin, 0, self.fradius//4, 1, digits = 0)
    self.connect_update_request(self.widgets.marginspin, "value-changed")
    wbox.pack(self.widgets.marginspin.hbox(prepend = "Frame margin:", append = "pixels"))
    self.widgets.fadespin = SpinButton(25, 0, 50, 1, digits = 1)
    self.connect_update_request(self.widgets.fadespin, "value-changed")
    wbox.pack(self.widgets.fadespin.hbox(prepend = "Fade length:", append = "% frame radius"))
    self.widgets.scalespin = SpinButton(1., .25, 4., .01, digits = 3)
    self.connect_update_request(self.widgets.scalespin, "value-changed")
    self.widgets.sizelabel = Label(" (0x0) px")
    wbox.pack(self.widgets.scalespin.hbox(prepend = "Image scale:", append = self.widgets.sizelabel))
    frame, hbox = FramedHBox(" Position ", margin = 32)
    wbox.pack(frame)
    grid = Grid(column_spacing = 0, column_homogeneous = True, row_spacing = 0, row_homogeneous = True)
    hbox.pack(grid, expand = True)
    self.widgets.cbutton = Button(label = "\u2022")
    self.widgets.cbutton.connect("clicked", lambda button: self.center_image())
    grid.attach(self.widgets.cbutton, 1, 1)
    self.widgets.ubutton = HoldButton(delay = self.delay)
    self.widgets.ubutton.add(Gtk.Arrow(arrow_type = Gtk.ArrowType.UP, shadow_type = Gtk.ShadowType.NONE))
    self.widgets.ubutton.connect("clicked", lambda button: self.move_image(0, +1))
    self.widgets.ubutton.connect("hold", lambda button: self.move_image(0, +10))
    grid.attach(self.widgets.ubutton, 1, 0)
    self.widgets.dbutton = HoldButton(delay = self.delay)
    self.widgets.dbutton.add(Gtk.Arrow(arrow_type = Gtk.ArrowType.DOWN, shadow_type = Gtk.ShadowType.NONE))
    self.widgets.dbutton.connect("clicked", lambda button: self.move_image(0, -1))
    self.widgets.dbutton.connect("hold", lambda button: self.move_image(0, -10))
    grid.attach(self.widgets.dbutton, 1, 2)
    self.widgets.lbutton = HoldButton(delay = self.delay)
    self.widgets.lbutton.add(Gtk.Arrow(arrow_type = Gtk.ArrowType.LEFT, shadow_type = Gtk.ShadowType.NONE))
    self.widgets.lbutton.connect("clicked", lambda button: self.move_image(-1, 0))
    self.widgets.lbutton.connect("hold", lambda button: self.move_image(-10, 0))
    grid.attach(self.widgets.lbutton, 0, 1)
    self.widgets.rbutton = HoldButton(delay = self.delay)
    self.widgets.rbutton.add(Gtk.Arrow(arrow_type = Gtk.ArrowType.RIGHT, shadow_type = Gtk.ShadowType.NONE))
    self.widgets.rbutton.connect("clicked", lambda button: self.move_image(+1, 0))
    self.widgets.rbutton.connect("hold", lambda button: self.move_image(+10, 0))
    grid.attach(self.widgets.rbutton, 2, 1)
    self.widgets.gbutton = CheckButton(label = "Show guide lines")
    self.widgets.gbutton.connect("toggled", lambda button: self.update_guide_lines(self.get_params()))
    wbox.pack(self.widgets.gbutton)
    wbox.pack(self.tool_control_buttons(model = "onthefly"))
    self.start(identity = False)
    return True

  def center_image(self):
    """Center image."""
    self.xcenter = 0
    self.ycenter = 0
    self.apply_idle()

  def move_image(self, dx, dy):
    """Move image by dx pixels along x and dy pixels along y."""
    self.xcenter += dx
    self.ycenter += dy
    self.apply_idle()

  def frame_mask(self, radius, margin, fade):
    """Return the mask for blending the image within the frame.
       'radius' is the frame radius (pixels), 'margin' the frame margin (pixels), and 'fade' the fade length (as a % of radius)."""
    x = np.arange(0, self.fwidth)-(self.fwidth-1)/2
    y = np.arange(0, self.fheight)-(self.fheight-1)/2
    X, Y = np.meshgrid(x, y, sparse = True)
    r = np.sqrt(X**2+Y**2)
    r0 = radius-margin
    r1 = radius-margin-fade*radius/100.
    mask = np.clip(self.maxfade+(1.-self.maxfade)*(r0-r)/(r0-r1), self.maxfade, 1.) if r0 > r1 else np.ones_like(r)
    mask = np.where(r <= r0, mask, 0.)
    return imageprocessing.IMGTYPE(mask)

  def plot_guide_lines(self, ax, radius, margin, fade):
    """Plot guide lines in axes 'ax' of the main window.
       'radius' is the frame radius (pixels), 'margin' the frame margin (pixels), and 'fade' the fade length (as a % of radius)."""
    xlim = ax.get_xlim()
    ylim = ax.get_ylim()
    dx = abs(xlim[1]-xlim[0])
    dy = abs(ylim[1]-ylim[0])
    #if abs(dx-self.fwidth) > .5 or abs(dy-self.fheight) > .5: return # Dont't draw guidelines if the image does not match the frame size.
    xc = (xlim[0]+xlim[1])/2.
    yc = (ylim[0]+ylim[1])/2.
    ax.guidelines = []
    ax.guidelines.append(ax.axvline(xc, color = "yellow", linestyle = "-.", linewidth = 1.))
    ax.guidelines.append(ax.axhline(yc, color = "yellow", linestyle = "-.", linewidth = 1.))
    ax.guidelines.append(ax.add_patch(plt.Circle((xc, yc), radius-margin, color = "yellow", linestyle = ":", linewidth = 1., fill = False)))
    ax.guidelines.append(ax.add_patch(plt.Circle((xc, yc), radius-margin-fade*radius/100., color = "yellow", linestyle = ":", linewidth = 1., fill = False)))

  def get_params(self):
    """Return tool parameters."""
    return self.xcenter, self.ycenter, self.widgets.scalespin.get_value(), self.widgets.marginspin.get_value(), self.widgets.fadespin.get_value()

  def set_params(self, params):
    """Set tool parameters 'params'."""
    self.xcenter, self.ycenter, scale, margin, fade = params
    self.widgets.scalespin.set_value(scale)
    self.widgets.marginspin.set_value(margin)
    self.widgets.fadespin.set_value(fade)

  def update_guide_lines(self, params, redraw = True):
    """Update guide lines in main window for parameters 'params'.
       The main window canvas is redrawn if redraw is True."""
    if self.widgets.gbutton.get_active():
      xc, yc, scale, margin, fade = params
      self.app.mainwindow.set_guide_lines(lambda ax: self.plot_guide_lines(ax, self.fradius, margin, fade), redraw)
    else:
      self.app.mainwindow.set_guide_lines(None, redraw)

  def run(self, params):
    """Run tool for parameters 'params'."""
    xcenter, ycenter, scale, margin, fade = params
    # Compute frame mask if needed.
    if (margin, fade) != self.currentmarginfade:
      self.fmask = self.frame_mask(self.fradius, margin, fade)
      self.currentmarginfade = (margin, fade)
      self.update_guide_lines(params, redraw = False) # Will redraw later.
    # Rescale image if needed.
    if scale != self.currentscale:
      if scale == 1.:
        self.rescaled = self.reference.clone()
      else:
        self.rescaled = self.reference.rescale(scale, resample = imageprocessing.LANCZOS, inplace = False)
      self.rwidth, self.rheight = self.rescaled.size()
      self.currentscale = scale
      self.widgets.sizelabel.set_label(f" ({self.rwidth}x{self.rheight} px)")
      self.currentmove = None # Force move & crop image.
    # Move & crop image if needed.
    if (xcenter, ycenter) != self.currentmove:
      xcmin = (self.rwidth-self.fwidth)//2-xcenter
      xcmax = xcmin+self.fwidth
      ycmin = (self.rheight-self.fheight)//2+ycenter
      ycmax = ycmin+self.fheight
      xfmin = 0
      xfmax = self.fwidth
      yfmin = 0
      yfmax = self.fheight
      if xcmin < 0:
        dx = -xcmin
        xcmin += dx
        xfmin += dx
      if xcmax > self.rwidth:
        dx = self.rwidth-xcmax
        xcmax += dx
        xfmax += dx
      if ycmin < 0:
        dy = -ycmin
        ycmin += dy
        yfmin += dy
      if ycmax > self.rheight:
        dy = self.rheight-ycmax
        ycmax += dy
        yfmax += dy
      self.cropped = imageprocessing.black_image(self.fwidth, self.fheight)
      self.cropped.rgb[:, yfmin:yfmax, xfmin:xfmax] = self.rescaled.rgb[:, ycmin:ycmax, xcmin:xcmax]
      self.currentmove = (xcenter, ycenter)
    # Blend image with frame.
    self.image.rgb = self.fmask*self.cropped.rgb+(1.-self.fmask)*self.frame.rgb
    return params, True

  def operation(self, params):
    """Return tool operation string for parameters 'params'."""
    xcenter, ycenter, scale, margin, fade = params
    return f"AddUnistellarFrame(file = '{self.basename}', scale = {scale:.3f}, margin = {margin:.0f}px, fade = {fade/100.:.3f}R)"

  def cleanup(self):
    """Free memory on exit."""
    try:
      del self.frame
      del self.fmask
      del self.rescaled
      del self.cropped
    except:
      pass



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated.

"""Settings window."""

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
from .gtk.customwidgets import Align, HBox, VBox, FramedHBox, HButtonBox, Button, CheckButton, SpinButton
from .base import BaseWindow, Container

class SettingsWindow(BaseWindow):
  """Settings window class."""

  def open(self):
    """Open tool window with title 'title' for image 'image'."""
    if self.opened: return
    self.opened = True
    self.window = Gtk.Window(title = "Settings",
                             transient_for = self.app.mainwindow.window,
                             border_width = 16,
                             modal = True)
    self.window.connect("delete-event", self.close)
    self.widgets = Container()
    wbox = VBox()
    self.window.add(wbox)
    frame, hbox = FramedHBox(" Apply operations on the fly (disable if not responsive) ", spacing = 16)
    wbox.pack(frame)
    vbox = VBox(spacing = 0, homogeneous = True)
    hbox.pack(vbox)
    self.widgets.hotpixelsbutton = CheckButton(label = "Hot pixels")
    self.widgets.hotpixelsbutton.set_active(self.app.hotpixelsotf)
    vbox.pack(self.widgets.hotpixelsbutton)
    self.widgets.stretchbutton = CheckButton(label = "Stretch tools")
    self.widgets.stretchbutton.set_active(self.app.stretchotf)
    vbox.pack(self.widgets.stretchbutton)
    self.widgets.colorbutton = CheckButton(label = "Color balance/saturation/noise")
    self.widgets.colorbutton.set_active(self.app.colorotf)
    vbox.pack(self.widgets.colorbutton)
    self.widgets.blendbutton = CheckButton(label = "Blend images")
    self.widgets.blendbutton.set_active(self.app.blendotf)
    vbox.pack(self.widgets.blendbutton)
    vbox = VBox(spacing = 0, valign = Align.CENTER)
    hbox.pack(vbox)
    self.widgets.timespin = SpinButton(self.app.polltime, 100, 1000, 10, digits = 0)
    vbox.pack(self.widgets.timespin.hbox(prepend = "Poll time:", append = "ms"))
    hbox = HButtonBox()
    wbox.pack(hbox)
    self.widgets.applybutton = Button(label = "OK")
    self.widgets.applybutton.connect("clicked", self.apply)
    hbox.pack(self.widgets.applybutton)
    self.widgets.applybutton = Button(label = "Reset")
    self.widgets.applybutton.connect("clicked", self.reset)
    hbox.pack(self.widgets.applybutton)
    self.widgets.cancelbutton = Button(label = "Cancel")
    self.widgets.cancelbutton.connect("clicked", self.close)
    hbox.pack(self.widgets.cancelbutton)
    self.window.show_all()

  def reset(self, *args, **kwargs):
    """Reset settings."""
    if not self.opened: return
    settings = self.app.get_default_settings()
    self.widgets.hotpixelsbutton.set_active(settings["remove_hot_pixels_on_the_fly"])
    self.widgets.colorbutton.set_active(settings["colors_on_the_fly"])
    self.widgets.stretchbutton.set_active(settings["stretch_on_the_fly"])
    self.widgets.blendbutton.set_active(settings["blend_on_the_fly"])
    self.widgets.timespin.set_value(settings["poll_time"])

  def apply(self, *args, **kwargs):
    """Apply settings."""
    if not self.opened: return
    self.app.hotpixelsotf = self.widgets.hotpixelsbutton.get_active()
    self.app.colorotf = self.widgets.colorbutton.get_active()
    self.app.stretchotf = self.widgets.stretchbutton.get_active()
    self.app.blendotf = self.widgets.blendbutton.get_active()
    self.app.polltime = int(self.widgets.timespin.get_value())
    self.app.save_settings() # Save current settings.
    self.close()

  def close(self, *args, **kwargs):
    """Close settings window."""
    if not self.opened: return
    self.window.destroy()
    self.opened = False
    del self.widgets



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated (+).

"""Gaussian filter tool."""

from ..gtk.customwidgets import VBox, HScaleSpinButton, RadioButtons
from ..toolmanager import BaseToolWindow
from skimage.filters import gaussian

class GaussianFilterTool(BaseToolWindow):
  """Gaussian filter tool class."""

  _action_ = "Applying gaussian filter..."

  _help_ = """Convolve the image with a gaussian of standard deviation \u03c3.
The image is extended across its boundaries according to the boundary mode:
  \u2022 Reflect: the image is reflected about the edge of the last pixel (abcd \u2192 dcba|abcd|dcba).
  \u2022 Mirror: the image is reflected about the center of the last pixel (abcd \u2192 dcb|abcd|cba).
  \u2022 Nearest: the image is padded with the value of the last pixel (abcd \u2192 aaaa|abcd|dddd).
  \u2022 Zero: the image is padded with zeros (abcd \u2192 0000|abcd|0000)."""

  _onthefly_ = False # This transformation can not be applied on the fly.

  def open(self, image):
    """Open tool window for image 'image'."""
    if not super().open(image, "Gaussian filter"): return False
    wbox = VBox()
    self.window.add(wbox)
    self.widgets.sigmascale = HScaleSpinButton(5., 0., 20., .01, digits = 2, length = 480)
    wbox.pack(self.widgets.sigmascale.layout2("\u03c3 (pixels):"))
    self.widgets.modebuttons = RadioButtons(("reflect", "Reflect"), ("mirror", "Mirror"), ("nearest", "Nearest"), ("zero", "Zero"))
    wbox.pack(self.widgets.modebuttons.hbox(prepend = "Boundary mode:"))
    wbox.pack(self.tool_control_buttons())
    self.start(identity = False)
    return True

  def get_params(self):
    """Return tool parameters."""
    return self.widgets.sigmascale.get_value(), self.widgets.modebuttons.get_selected()

  def set_params(self, params):
    """Set tool parameters 'params'."""
    sigma, mode = params
    self.widgets.sigmascale.set_value(sigma)
    self.widgets.modebuttons.set_selected(mode)

  def run(self, params):
    """Run tool for parameters 'params'."""
    sigma, mode = params
    if sigma <= 0: return params, False
    if mode == "zero": # Translate modes for denoise_bilateral.
      mode = "constant"
    self.image.rgb = gaussian(self.reference.rgb, channel_axis = 0, sigma = sigma, mode = mode, cval = 0.)
    return params, True

  def operation(self, params):
    """Return tool operation string for parameters 'params'."""
    sigma, mode = params
    return f"GaussianFilter(sigma = {sigma:.2f} pixels, mode = {mode})"



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17

"""Image processing utils."""

import numpy as np
from .defs import IMGTYPE, IMGTOL

#############################
# Generic image validation. #
#############################

def is_valid_rgb_image(image):
  """Return True if 'image' is a valid RGB image, False otherwise."""
  if not isinstance(image, np.ndarray): return False
  if image.ndim != 3: return False
  if image.shape[0] != 3: return False
  if image.dtype != IMGTYPE: return False
  return True

###############################
# Image manipulation helpers. #
###############################

def failsafe_divide(A, B):
  """Return A/B, ignoring errors (division by zero, ...)."""
  status = np.seterr(all = "ignore")
  C = np.divide(A, B)
  np.seterr(divide = status["divide"], over = status["over"], under = status["under"], invalid = status["invalid"])
  return C

def scale_pixels(image, source, target, cutoff = IMGTOL):
  """Scale all pixels of the image 'image' by the ratio target/source.
     Wherever abs(source) < cutoff, set all channels to target."""
  return np.where(abs(source) > cutoff, failsafe_divide(image*target, source), target)

def lookup(x, xlut, ylut, slut, nlut):
  """Return y = f(x) by linearly interpolating the values ylut = f(xlut) of an evenly spaced look-up table with nlut elements.
     slut = (ylut[1:]-ylut[:-1])/(xlut[1:]-xlut[:-1]) are the slopes used for linear interpolation between successive elements."""
  l = np.clip(np.int32(np.floor((x-xlut[0])*(nlut-1)/(xlut[-1]-xlut[0]))), 0, nlut-2)
  return slut[l]*(x-xlut[l])+ylut[l]



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated.

"""Application menus."""

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Gio
from .gtk.customwidgets import CheckButton
from .gtk.filechoosers import ImageFileChooserDialog
from .base import ErrorDialog
from .settings import SettingsWindow
from .tools.blackpoint import BlackPointTool
from .tools.arcsinh import ArcsinhStretchTool
from .tools.hyperbolic import GeneralizedHyperbolicStretchTool
from .tools.midtone import MidtoneStretchTool
from .tools.clahe import CLAHETool
from .tools.colorbalance import ColorBalanceTool
from .tools.colorsaturation import ColorSaturationTool
from .tools.ghscolorsat import GHSColorSaturationTool
from .tools.colornoise import ColorNoiseReductionTool
from .tools.grayscale import GrayScaleConversionTool
from .tools.hotpixels import RemoveHotPixelsTool
from .tools.gaussian import GaussianFilterTool
from .tools.butterworth import ButterworthFilterTool
from .tools.wavelets import WaveletsFilterTool
from .tools.nlmeans import NonLocalMeansFilterTool
from .tools.bilateral import BilateralFilterTool
from .tools.totalvariation import TotalVariationFilterTool
from .tools.unsharp import UnsharpMaskTool
from .tools.thresholdmask import ThresholdMaskTool
from .tools.blend import BlendTool
from .tools.pixelmath import PixelMathTool
from .tools.resample import ResampleTool
from .tools.addframe import AddUnistellarFrame
from .tools.switch import SwitchTool
from . import editors

XMLMENUS = """
<?xml version="1.0" encoding="UTF-8"?>
<interface>
  <menu id="MainMenu">
    <submenu>
      <attribute name="label">File</attribute>
      <section>
        <item>
          <attribute name="label">Open</attribute>
          <attribute name="action">app.open</attribute>
          <attribute name="accel">&lt;Primary&gt;o</attribute>
        </item>
        <item>
          <attribute name="label">Save</attribute>
          <attribute name="action">app.save</attribute>
          <attribute name="accel">&lt;Primary&gt;s</attribute>
        </item>
        <item>
          <attribute name="label">Close</attribute>
          <attribute name="action">app.close</attribute>
        </item>
      </section>
      <section>
        <item>
          <attribute name="label">Settings</attribute>
          <attribute name="action">app.settings</attribute>
        </item>
      </section>
      <section>
        <item>
          <attribute name="label">Quit</attribute>
          <attribute name="action">app.quit</attribute>
          <attribute name="accel">&lt;Primary&gt;q</attribute>
        </item>
      </section>
    </submenu>
    <submenu>
      <attribute name="label">Stretch</attribute>
      <section>
        <item>
          <attribute name="label">Black point</attribute>
          <attribute name="action">app.blackpoint</attribute>
        </item>
      </section>
      <section>
        <item>
          <attribute name="label">Arcsinh stretch</attribute>
          <attribute name="action">app.asinhstretch</attribute>
        </item>
        <item>
          <attribute name="label">Generalized hyperbolic stretch</attribute>
          <attribute name="action">app.GHstretch</attribute>
        </item>
        <item>
          <attribute name="label">Midtone stretch</attribute>
          <attribute name="action">app.MTstretch</attribute>
        </item>
      </section>
      <section>
        <item>
          <attribute name="label">CLAHE</attribute>
          <attribute name="action">app.CLAHE</attribute>
        </item>
      </section>
    </submenu>
    <submenu>
      <attribute name="label">Colors</attribute>
      <section>
        <item>
          <attribute name="label">Color balance</attribute>
          <attribute name="action">app.colorbalance</attribute>
        </item>
        <item>
          <attribute name="label">Color saturation</attribute>
          <attribute name="action">app.colorsaturation</attribute>
        </item>
        <item>
          <attribute name="label">Color saturation hyperbolic stretch</attribute>
          <attribute name="action">app.GHScolorsat</attribute>
        </item>
        <item>
          <attribute name="label">Color noise reduction</attribute>
          <attribute name="action">app.colornoise</attribute>
        </item>
      </section>
      <section>
        <item>
          <attribute name="label">Negative</attribute>
          <attribute name="action">app.negative</attribute>
        </item>
        <item>
          <attribute name="label">Convert to gray scale</attribute>
          <attribute name="action">app.grayscale</attribute>
        </item>
      </section>
      <section>
        <item>
          <attribute name="label">Convert from lRGB to sRGB</attribute>
          <attribute name="action">app.lRGBtosRGB</attribute>
        </item>
      </section>
    </submenu>
    <submenu>
      <attribute name="label">Filters</attribute>
      <section>
        <item>
          <attribute name="label">Remove hot pixels</attribute>
          <attribute name="action">app.hotpixels</attribute>
        </item>
      </section>
      <section>
        <item>
          <attribute name="label">Gaussian filter</attribute>
          <attribute name="action">app.gaussian</attribute>
        </item>
        <item>
          <attribute name="label">Butterworth filter</attribute>
          <attribute name="action">app.butterworth</attribute>
        </item>
        <item>
          <attribute name="label">Wavelets filter</attribute>
          <attribute name="action">app.wavelets</attribute>
        </item>
      </section>
      <section>
        <item>
          <attribute name="label">Non-local means filter</attribute>
          <attribute name="action">app.nlmeans</attribute>
        </item>
        <item>
          <attribute name="label">Bilateral filter</attribute>
          <attribute name="action">app.bilateral</attribute>
        </item>
        <item>
          <attribute name="label">Total variation filter</attribute>
          <attribute name="action">app.totalvariation</attribute>
        </item>
      </section>
      <section>
        <item>
          <attribute name="label">Sharpen (Laplacian)</attribute>
          <attribute name="action">app.sharpen</attribute>
        </item>
        <item>
          <attribute name="label">Unsharp mask</attribute>
          <attribute name="action">app.unsharp</attribute>
        </item>
      </section>
    </submenu>
    <submenu>
      <attribute name="label">Masks</attribute>
      <section>
        <item>
          <attribute name="label">Threshold mask</attribute>
          <attribute name="action">app.thresholdmask</attribute>
        </item>
      </section>
    </submenu>
    <submenu>
      <attribute name="label">Operations</attribute>
      <section>
        <item>
          <attribute name="label">Blend</attribute>
          <attribute name="action">app.blend</attribute>
        </item>
      </section>
      <section>
        <item>
          <attribute name="label">Pixel math</attribute>
          <attribute name="action">app.pixelmath</attribute>
        </item>
      </section>
      <section>
        <item>
          <attribute name="label">Resample</attribute>
          <attribute name="action">app.resample</attribute>
        </item>
      </section>
      <section>
        <item>
          <attribute name="label">Edit with siril</attribute>
          <attribute name="action">app.siril</attribute>
        </item>
        <item>
          <attribute name="label">Edit with gimp</attribute>
          <attribute name="action">app.gimp</attribute>
        </item>
        <item>
          <attribute name="label">Edit with ...</attribute>
          <attribute name="action">app.edit</attribute>
        </item>
      </section>
    </submenu>
    <submenu>
      <attribute name="label">Frames</attribute>
      <section>
        <item>
          <attribute name="label">Remove frame</attribute>
          <attribute name="action">app.removeframe</attribute>
        </item>
        <item>
          <attribute name="label">Restore frame</attribute>
          <attribute name="action">app.restoreframe</attribute>
        </item>
      </section>
      <section>
        <item>
          <attribute name="label">Add frame</attribute>
          <attribute name="action">app.addframe</attribute>
        </item>
      </section>
    </submenu>
    <submenu>
      <attribute name="label">Logs</attribute>
      <section>
        <item>
          <attribute name="label">Undo last operation</attribute>
          <attribute name="action">app.undo</attribute>
          <attribute name="accel">&lt;Primary&gt;z</attribute>
        </item>
        <item>
          <attribute name="label">Redo last operation</attribute>
          <attribute name="action">app.redo</attribute>
          <attribute name="accel">&lt;Primary&gt;r</attribute>
        </item>
      </section>
      <section>
        <item>
          <attribute name="label">Switch to an other image</attribute>
          <attribute name="action">app.switch</attribute>
        </item>
      </section>
      <section>
        <item>
          <attribute name="label">View logs</attribute>
          <attribute name="action">app.viewlogs</attribute>
          <attribute name="accel">&lt;Primary&gt;l</attribute>
        </item>
      </section>
    </submenu>
  </menu>
  <menu id="MainWindowContextMenu">
    <item>
      <attribute name="label">Statistics</attribute>
      <attribute name="action">app.statistics</attribute>
    </item>
    <item>
      <attribute name="label">Light curve</attribute>
      <attribute name="action">app.lightcurve</attribute>
    </item>
  </menu>
</interface>
"""

builder = Gtk.Builder.new_from_string(XMLMENUS, -1)

class Actions:
  """Menu actions class."""

  def __init__(self, app):
    """Set-up menu actions for application 'app'.
       All actions are attached to app for simplicity."""

    def add_action(name, callback, context = {}):
      """Add action with name 'name', callback 'callback', and context modifiers 'context'
         (with respect to the default context {"noimage": False, "activetool": False, "nocancelled": True})."""
      action = Gio.SimpleAction.new(name, None)
      action.connect("activate", callback)
      app.add_action(action)
      actioncontext = {"noimage": False, "activetool": False, "nocancelled": True}
      actioncontext.update(context)
      self.actions.append((action, actioncontext))

    self.app = app
    self.actions = []
    #
    ######################
    # Main menu actions. #
    ######################
    #
    ### File.
    #
    add_action("open", self.load_file, {"noimage": True})
    add_action("save", self.save_file)
    add_action("close", self.close)
    #
    add_action("settings", lambda action, parameter: SettingsWindow(app).open(), {"noimage": True})
    #
    add_action("quit", lambda action, parameter: app.mainwindow.close(), {"noimage": True, "activetool": True})
    #
    ### Stretch.
    #
    add_action("blackpoint", lambda action, parameter: app.run_tool(BlackPointTool, app.stretchotf))
    #
    add_action("asinhstretch", lambda action, parameter: app.run_tool(ArcsinhStretchTool, app.stretchotf))
    add_action("GHstretch", lambda action, parameter: app.run_tool(GeneralizedHyperbolicStretchTool, app.stretchotf))
    add_action("MTstretch", lambda action, parameter: app.run_tool(MidtoneStretchTool, app.stretchotf))
    #
    add_action("CLAHE", lambda action, parameter: app.run_tool(CLAHETool))
    #
    ### Colors.
    #
    add_action("colorbalance", lambda action, parameter: app.run_tool(ColorBalanceTool, app.colorotf))
    add_action("colorsaturation", lambda action, parameter: app.run_tool(ColorSaturationTool, app.colorotf))
    add_action("GHScolorsat", lambda action, parameter: app.run_tool(GHSColorSaturationTool, app.stretchotf))
    add_action("colornoise", lambda action, parameter: app.run_tool(ColorNoiseReductionTool, app.colorotf))
    #
    add_action("negative", lambda action, parameter: app.negative())
    add_action("grayscale", lambda action, parameter: app.run_tool(GrayScaleConversionTool, app.colorotf))
    #
    add_action("lRGBtosRGB", lambda action, parameter: app.lrgb_to_srgb())
    #
    ### Filters.
    #
    add_action("hotpixels", lambda action, parameter: app.run_tool(RemoveHotPixelsTool, app.hotpixelsotf))
    #
    add_action("gaussian", lambda action, parameter: app.run_tool(GaussianFilterTool))
    add_action("butterworth", lambda action, parameter: app.run_tool(ButterworthFilterTool))
    add_action("wavelets", lambda action, parameter: app.run_tool(WaveletsFilterTool))
    add_action("bilateral", lambda action, parameter: app.run_tool(BilateralFilterTool))
    add_action("nlmeans", lambda action, parameter: app.run_tool(NonLocalMeansFilterTool))
    add_action("totalvariation", lambda action, parameter: app.run_tool(TotalVariationFilterTool))
    #
    add_action("sharpen", lambda action, parameter: app.sharpen())
    add_action("unsharp", lambda action, parameter: app.run_tool(UnsharpMaskTool))
    #
    ### Masks.
    #
    add_action("thresholdmask", lambda action, parameter: app.run_tool(ThresholdMaskTool))
    #
    ### Operations.
    #
    add_action("blend", lambda action, parameter: app.run_tool(BlendTool, app.blendotf))
    #
    add_action("pixelmath", lambda action, parameter: app.run_tool(PixelMathTool))
    #
    add_action("resample", lambda action, parameter: app.run_tool(ResampleTool))
    #
    add_action("siril", lambda action, parameter: editors.edit_with_siril(app))
    add_action("gimp", lambda action, parameter: editors.edit_with_gimp(app))
    add_action("edit", lambda action, parameter: editors.edit_with_any(app))
    #
    ### Frames.
    #
    add_action("removeframe", lambda action, parameter: app.remove_unistellar_frame())
    add_action("restoreframe", lambda action, parameter: app.restore_unistellar_frame())
    #
    add_action("addframe", lambda action, parameter: app.run_tool(AddUnistellarFrame))
    #
    ### Logs.
    #
    add_action("undo", lambda action, parameter: app.undo())
    add_action("redo", lambda action, parameter: app.redo(), {"noimage": True, "nocancelled": False})
    #
    add_action("switch", lambda action, parameter: app.run_tool(SwitchTool))
    #
    add_action("viewlogs", lambda action, parameter: app.logwindow.open(), {"activetool": True})
    #
    #####################################
    # Main window context menu actions. #
    #####################################
    #
    add_action("statistics", lambda action, parameter: app.mainwindow.show_statistics(), {"noimage": True, "activetool": True})
    add_action("lightcurve", lambda action, parameter: app.mainwindow.show_lightcurve(), {"activetool": True})
    #
    ###
    #
    self.update()

  def update(self):
    """Update menu actions according to the application context."""
    context = self.app.get_context()
    for action, enable in self.actions:
      if not context["image"]:
        enabled = enable["noimage"]
      elif context["activetool"]:
        enabled = enable["activetool"]
      else:
        enabled = True
      if not context["cancelled"]:
        enabled = enabled and enable["nocancelled"]
      action.set_enabled(enabled)

  def load_file(self, *args, **kwargs):
    """Open file dialog and load image file."""
    filename = ImageFileChooserDialog(self.app.mainwindow.window, Gtk.FileChooserAction.OPEN, preview = True)
    if filename is None: return
    try:
      self.app.load_file(filename)
    except Exception as err:
      ErrorDialog(self.app.mainwindow.window, str(err))

  def save_file(self, *args, **kwargs):
    """Open file dialog and save image file."""
    if not self.app.get_context("image"): return
    # Add extra widget to choose the color depth of png and tiff files.
    depthbutton = CheckButton(label = "16 bits color depth (for png and tiff files)")
    filename = ImageFileChooserDialog(self.app.mainwindow.window, Gtk.FileChooserAction.SAVE, path = self.app.get_savename(), extra_widget = depthbutton)
    if filename is None: return
    try:
      self.app.save_file(filename, depth = 16 if depthbutton.get_active() else 8)
    except Exception as err:
      ErrorDialog(self.app.mainwindow.window, str(err))

  def close(self, *args, **kwargs):
    """Close current image."""
    if not self.app.get_context("image"): return
    dialog = Gtk.MessageDialog(transient_for = self.app.mainwindow.window,
                               message_type = Gtk.MessageType.QUESTION,
                               buttons = Gtk.ButtonsType.OK_CANCEL,
                               modal = True)
    dialog.set_markup("Are you sure you want to close this image ?")
    response = dialog.run()
    dialog.destroy()
    if response != Gtk.ResponseType.OK: return True
    self.app.clear()



#!/usr/bin/python

# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17

"""eQuimageGtk is a python tool (with a Gtk GUI) to postprocess astronomical images from Unistellar telescopes."""

__version__ = "1.8.0"

import os
os.environ["LANGUAGE"] = "en"
import sys
import ast
import inspect
packagepath = os.path.dirname(inspect.getabsfile(inspect.currentframe()))
import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Gdk, Gio
import matplotlib.pyplot as plt
plt.style.use(os.path.join(packagepath, "config", "eQuimage.mplstyle"))
from .gui import menus
from .gui.base import ErrorDialog
from .gui.mainwindow import MainWindow
from .gui.toolmanager import BaseToolWindow
from .gui.logs import LogWindow
from .imageprocessing import imageprocessing
from .imageprocessing.Unistellar import UnistellarImage

class eQuimageGtkApp(Gtk.Application):
  """The eQuimageGtk application."""

  #######################
  # Gtk initialization. #
  #######################

  def __init__(self, *args, **kwargs):
    """Initialize the eQuimageGtk application."""
    super().__init__(*args, flags = Gio.ApplicationFlags.HANDLES_OPEN, **kwargs)
    self.initialize()

  def do_startup(self):
    """Load CSS and prepare main menu on startup."""
    Gtk.Application.do_startup(self)
    # Load CSS.
    screen = Gdk.Screen.get_default()
    provider = Gtk.CssProvider()
    provider.load_from_path(os.path.join(self.packagepath, "config", "eQuimage.css"))
    stylecontext = Gtk.StyleContext()
    stylecontext.add_provider_for_screen(screen, provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
    # Set-up actions and main menu.
    self.actions = menus.Actions(self)
    self.set_menubar(menus.builder.get_object("MainMenu"))

  def do_activate(self):
    """Open the main window on activation."""
    self.mainwindow.open()
    if imageprocessing.IMAGEIO:
      try: # Download freeimage plugin for imageio...
        import imageio
        imageio.plugins.freeimage.download()
      except:
        ErrorDialog(self.mainwindow.window, "Failed to download and install the freeimage plugin for imageio.")
        self.quit()

  def do_open(self, files, nfiles, hint):
    """Open command line file."""
    if nfiles > 1:
      print("Syntax : eQuimageGtk [image_file] where [image_file] is the image to open.")
      self.quit()
    self.activate()
    try:
      self.load_file(files[0].get_path())
    except Exception as err:
      ErrorDialog(self.mainwindow.window, str(err))

  ###############################
  # Application data & methods. #
  ###############################

  ImageClass = UnistellarImage # The base class used for images.

  # Initialization.

  def initialize(self):
    """Initialize the eQuimageGtkApp object."""
    self.version = __version__
    self.packagepath = packagepath
    self.mainwindow = MainWindow(self)
    self.toolwindow = BaseToolWindow(self)
    self.logwindow  = LogWindow(self)
    self.default_settings()
    if self.load_settings() > 0: self.save_settings() # Overwrite invalid or incomplete configuration file.
    self.reset()

  def reset(self):
    """Reset the eQuimageGtkApp object data."""
    self.filename = None
    self.pathname = None
    self.basename = None
    self.savename = None
    self.frame = None
    self.images = []
    self.operations = []
    self.cancelled = []
    self.width = 0
    self.height = 0
    self.colordepth = 8
    self.meta = {}

  def clear(self, mainwindow = True):
    """Clear the eQuimageGtkApp object data and windows.
       Don't clear the main window if mainwindow is False."""
    if self.filename is not None: print(f"Closing {self.filename}...")
    self.reset()
    self.logwindow.close()
    self.toolwindow.destroy()
    if mainwindow: self.mainwindow.reset_images()
    self.actions.update()

  # Application context.

  def get_context(self, key = None):
    """Return the application context:
         - get_context("image") = True if an image is loaded.
         - get_context("activetool") = True if a tool is active.
         - get_context("cancelled") = True if there are cancelled operations available for redo.
         - get_context() returns all above keys as a dictionnary."""
    context = {"image": len(self.images) > 0, "activetool": self.toolwindow.opened, "cancelled": len(self.cancelled) > 0}
    return context[key] if key is not None else context

  def get_image_size(self):
    """Return width and height of the *original* image."""
    return self.width, self.height

  def get_color_depth(self):
    """Return color depth (bits per channel)."""
    return self.colordepth

  # File management.

  def get_packagepath(self):
    """Return package path."""
    return self.packagepath

  def get_filename(self):
    """Return image file name."""
    return self.filename

  def get_basename(self):
    """Return image base name."""
    return self.basename

  def get_pathname(self):
    """Return image path name."""
    return self.pathname

  def get_savename(self):
    """Return image save name."""
    return self.savename

  def load_file(self, filename):
    """Load image file 'filename'."""
    image = self.ImageClass()
    meta = image.load(filename)
    if not image.is_valid(): return
    self.clear(mainwindow = False)
    self.meta = meta
    self.filename = filename
    self.pathname = os.path.dirname(filename)
    self.basename = os.path.basename(filename)
    root, ext = os.path.splitext(filename)
    self.savename = root+"-post"+ext
    self.width, self.height = image.size() # *Original* image size.
    self.colordepth = self.meta["colordepth"] # Bits per channel.
    framed = image.check_frame()
    if framed:
      print(f"""Image has a frame type '{framed["type"]}'.""")
      self.frame = image.get_frame()
    else:
      self.frame = None
    image.meta["description"] = "Original image"
    self.push_operation(f"Load('{self.basename}')", image, self.frame)
    self.mainwindow.reset_images()
    self.actions.update()

  def save_file(self, filename = None, depth = 8):
    """Save image in file 'filename' (defaults to self.savename if None) with color depth 'depth' (bits/channel)."""
    if not self.images: return
    if filename is None: filename = self.savename
    self.images[-1].save(filename, depth = depth)
    root, ext = os.path.splitext(filename)
    with open(root+".log", "w") as f: f.write(self.logs())
    self.savename = filename

  # Images stack.

  def push_image(self, image, clone = False):
    """Push (or clone) image 'image' on top of the images stack."""
    self.images.append(image.clone() if clone else image)
    if self.images[-1] is not None:
      if self.images[-1].rgb.dtype != imageprocessing.IMGTYPE:
        print(f"Warning: The last image pushed on the stack is not of type {str(imageprocessing.IMGTYPE)}.")
    return self.images[-1]

  def pop_image(self):
    """Pop and return image from the top of the images stack."""
    return self.images.pop()

  def get_nbr_images(self):
    """Return the number of images in the images stack."""
    return len(self.images)

  def get_image(self, index = -1):
    """Return image with index 'index' from the images stack."""
    return self.images[index]

  # Operations stack.

  def push_operation(self, operation, image, frame = None):
    """Push operation 'operation' on image 'image' on top of the operations and images stacks.
       If not None, 'frame' is the new image frame."""
    if frame is not None:
      self.frame = self.push_image(frame, clone = True)
    else:
      self.push_image(self.frame) # Just push a reference; do not duplicate the current frame.
    self.push_image(image, clone = True)
    self.operations.append((operation, self.images[-1], self.images[-2]))

  def pop_operation(self):
    """Pop and return last (operation, image, frame) from the top of the operations stack.
       The images stack is truncated accordingly."""
    operation, image, frame = self.operations.pop()
    index = self.images.index(image)
    self.images = self.images[:index-1] # Include frame.
    return operation, image, frame

  def get_nbr_operations(self):
    """Return the number of operations in the operations stack."""
    return len(self.operations)

  def logs(self):
    """Return logs from the operations stack."""
    text = f"eQuimageGtk v{self.version}\n"
    n = 0
    for operation, *images in self.operations:
      n += 1
      text += f"IMG{n} = {operation}\n"
    return text

  # Tools management.

  def run_tool(self, ToolClass, onthefly = True):
    """Run tool 'ToolClass'.
       Apply tool and update the main window on the fly if 'onthefly' is True."""
    if self.toolwindow.opened: return
    self.toolwindow = ToolClass(self, self.polltime if onthefly else -1)
    if not self.toolwindow.open(self.images[-1]): return
    self.actions.update()
    self.toolwindow.window.present()

  def finalize_tool(self, image, operation, frame = None):
    """Finalize tool: push ('operation', 'image', 'frame') on the operations and images stacks (if operation is not None)
       and refresh main menu, main window, and log window.
       If 'frame' is None, the current self.frame is used as image frame from now on."""
    if operation is not None:
      image.meta["description"] = operation
      self.push_operation(operation, image, frame)
      self.cancelled = []
    self.mainwindow.reset_images()
    self.logwindow.update()
    self.actions.update()

  def undo(self):
    """Cancel last operation."""
    if self.toolwindow.opened: return
    if not self.operations: return
    print("Cancelling last operation...")
    self.cancelled.append(self.pop_operation())
    if self.images: self.frame = self.images[-2] # Update current frame.
    self.mainwindow.reset_images()
    self.logwindow.update()
    self.actions.update()

  def redo(self):
    """Redo last operation."""
    if self.toolwindow.opened: return
    if not self.cancelled: return
    print("Redoing last operation...")
    self.push_operation(*self.cancelled.pop())
    self.mainwindow.reset_images()
    self.logwindow.update()
    self.actions.update()

  # Simple tools.

  def sharpen(self):
    """Sharpen image."""
    if self.toolwindow.opened: return
    print("Sharpening image (with Laplacian kernel)...")
    self.finalize_tool(self.images[-1].sharpen(inplace = False), "Sharpen()")

  def negative(self):
    """Make a negative of the image."""
    if self.toolwindow.opened: return
    print("Converting to negative...")
    self.finalize_tool(self.images[-1].negative(inplace = False), "Negative()")

  def lrgb_to_srgb(self):
    """Convert the image from linear to sRGB color space."""
    if self.toolwindow.opened: return
    print("Converting from lRGB to sRGB...")
    self.finalize_tool(self.ImageClass(self.images[-1].lrgb_to_srgb(), self.images[-1].meta), "lRGBtosRGB()")

  def remove_unistellar_frame(self):
    """Remove Unistellar frame."""
    if self.toolwindow.opened: return
    frame = self.images[-1].get_frame()
    if frame is None:
      ErrorDialog(self.mainwindow.window, "This image has no frame.")
      return
    print("Removing Unistellar frame...")
    self.finalize_tool(self.images[-1].remove_frame(frame, inplace = False), "RemoveUnistellarFrame()")

  def restore_unistellar_frame(self):
    """Restore Unistellar frame."""
    if self.toolwindow.opened: return
    if self.frame is None:
      ErrorDialog(self.mainwindow.window, "This is no registered frame.")
      return
    print("Restoring Unistellar frame...")
    try:
      image = self.images[-1].add_frame(self.frame, inplace = False)
    except:
      ErrorDialog(self.mainwindow.window, "Operation failed.")
      return
    self.finalize_tool(image, "RestoreUnistellarFrame()")

  # Settings.

  def settings_from_dict(self, settings):
    """Set settings from dictionnary 'settings'.
       Return zero if successful, non-zero otherwise."""
    error = 0
    try: # Apply remove hot pixels tool on the fly ?
      self.hotpixelsotf = bool(settings["remove_hot_pixels_on_the_fly"])
    except:
      print("remove_hot_pixels_on_the_fly keyword not found in configuration file.")
      error = 1
    try: # Apply stretch tools on the fly ?
      self.stretchotf = bool(settings["stretch_on_the_fly"])
    except:
      print("stretch_on_the_fly keyword not found in configuration file.")
      error = 2
    try: # Apply color tools on the fly ?
      self.colorotf = bool(settings["colors_on_the_fly"])
    except:
      print("colors_on_the_fly keyword not found in configuration file.")
      error = 3
    try: # Apply blend tool on the fly ?
      self.blendotf = bool(settings["blend_on_the_fly"])
    except:
      print("blend_on_the_fly keyword not found in configuration file.")
      error = 4
    try: # Poll for new operations every self.polltime ms.
      self.polltime = int(settings["poll_time"])
    except:
      print("poll_time keyword not found in configuration file.")
      error = 5
    return error

  def get_default_settings(self):
    """Return default settings as a dictionnary."""
    return {"remove_hot_pixels_on_the_fly": True, "stretch_on_the_fly": True, "colors_on_the_fly": True, "blend_on_the_fly": True, "poll_time": 333}

  def default_settings(self):
    """Apply default settings."""
    self.settings_from_dict(self.get_default_settings())

  def load_settings(self):
    """Read settings in (system wide) file packagepath/eQuimagerc.
       Return zero if successful, non-zero otherwise."""
    filename = os.path.join(self.packagepath, "config", "eQuimagerc")
    try:
      with open(filename, "r") as f:
        string = f.readline()
      settings = ast.literal_eval(string)
      if not isinstance(settings, dict): raise TypeError
    except:
      print(f"Failed to read configuration file {filename}.")
      return -1
    return self.settings_from_dict(settings)

  def save_settings(self):
    """Save settings in (system wide) file packagepath/eQuimagerc.
       Return zero if successful, non-zero otherwise."""
    error = 0
    settings = {"remove_hot_pixels_on_the_fly": self.hotpixelsotf, "stretch_on_the_fly": self.stretchotf, "colors_on_the_fly": self.colorotf, "blend_on_the_fly": self.blendotf, "poll_time": self.polltime}
    filename = os.path.join(self.packagepath, "config", "eQuimagerc")
    try:
      with open(filename, "w") as f:
        f.write(repr(settings))
    except:
      print(f"Failed to write configuration file {filename}.")
      error = -1
    return error

#

def run():
  """Run eQuimageGtk."""
  application = eQuimageGtkApp()
  application.run(sys.argv)

#

if __name__ == "__main__": run()
